from offsetbasedgraph.tracevariants import *
import matplotlib.pyplot as plt
import numpy as np
import sys
from scipy.stats import binom_test
import motif_models

count_mismatch = lambda result: result.total - (result.A_count + result.B_count)

def check_is_diplo(result):
    return result.total == result.A_count+result.B_count


def _get_binomial_ps(mismatches, total, rate):
    return np.array([binom_test(missees, n, p=rate, alternative="greater")
                     for missees, n in zip(mismatches, total)])


def get_motif_plot(motif_results, all_results):
    motif_mismatches = np.array([count_mismatch(result) for result in motif_results])
    all_mismatches = np.array([count_mismatch(result) for result in all_results])
    motif_total = np.array([result.total for result in motif_results])
    all_total = np.array([result.total for result in all_results])
    mismatch_rate = np.median(all_mismatches/all_total)
    threshold = 1
    motif_ps = np.where(motif_total>0, 1-motif_mismatches/motif_total, 0)
    all_ps = np.where(all_total>0, 1-all_mismatches/all_total, 0)
    motif_counts = np.cumsum(np.histogram(motif_ps[::-1], 100, (0, 1))[0])
    all_counts = np.cumsum(np.histogram(all_ps[::-1], 100, (0, 1))[0])
    sorted_all = np.sort(1-all_ps)
    sorted_motif = np.sort(1-motif_ps)
    motif_plot = np.digitize(sorted_all, sorted_motif)/(1+np.arange(sorted_all.size))
    idxs = np.r_[np.flatnonzero(np.diff(sorted_all)), motif_plot.size-1]
    motif_plot = motif_plot[idxs]
    return idxs, motif_plot
    # return sorted_all[idxs], motif_plot


def compare_two(motif_results_A, all_results_A, motif_results_B, all_results_B):
    motif_plot_A = get_motif_plot(motif_results_A, all_results_A)
    motif_plot_B = get_motif_plot(motif_results_B, all_results_B)
    # args = np.argsort(np.r_[motif_plot_A[0], motif_plot_B[0]])
    # B_idxs = args >= motif_plot_A[0].size
    # A_idxs = ~B_idxs
    # B_idxs = np.flatnonzero(B_idxs)
    # A_idxs = np.flatnonzero(A_idxs)
    A_idxs = motif_plot_A[0]
    B_idxs = motif_plot_B[0]
    print(A_idxs.shape, motif_plot_A[1].shape)
    print(B_idxs.shape, motif_plot_B[1].shape)
    plt.plot(A_idxs, motif_plot_A[1], label="graph")
    plt.plot(B_idxs, motif_plot_B[1], label="macs")
    plt.legend()
    # plt.show()
    # plt.plot(motif_plot_A[1], label="graph")
    # plt.plot(motif_plot_B[1], label="macs")

def analyze_result(motif_results, all_results, col="b"):
    motif_mismatches = np.array([count_mismatch(result) for result in motif_results])
    all_mismatches = np.array([count_mismatch(result) for result in all_results])
    motif_total = np.array([result.total for result in motif_results])
    all_total = np.array([result.total for result in all_results])
    mismatch_rate = np.median(all_mismatches/all_total)
    threshold = 1
    print(mismatch_rate)
    # mismatch_rate = np.sum(all_mismatches)/np.sum(all_total)
    # print(mismatch_rate)
    motif_ps = np.where(motif_total>0, 1-motif_mismatches/motif_total, 0)
    all_ps = np.where(all_total>0, 1-all_mismatches/all_total, 0)
    # motif_ps = _get_binomial_ps(motif_mismatches, motif_total, mismatch_rate)
    # all_ps = _get_binomial_ps(all_mismatches, all_total, mismatch_rate)
    # plt.hist(_motif_ps, 20, (0, 1))
    # plt.show()
    # plt.hist(_all_ps, 20, (0, 1))
    # plt.show()
    motif_counts = np.cumsum(np.histogram(motif_ps[::-1], 100, (0, 1))[0])
    # plt.hist(motif_ps, 200)
    # plt.show()
    all_counts = np.cumsum(np.histogram(all_ps[::-1], 100, (0, 1))[0])
    # plt.hist(all_ps, 200)
    # plt.show()
    sorted_all = np.sort(1-all_ps)
    sorted_motif = np.sort(1-motif_ps)
    motif_plot = np.digitize(sorted_all, sorted_motif)/(1+np.arange(sorted_all.size))
    idxs = np.r_[np.flatnonzero(np.diff(sorted_all)), motif_plot.size-1]
    # split = np.flatnonzero(sorted_all[idxs] > (1-threshold))[0]
    motif_plot = motif_plot[idxs]
    # plt.plot(idxs[:split], motif_plot[:split], col)
    # plt.plot(idxs[split:], motif_plot[split:], col)
    plt.plot(idxs, motif_plot, col)
    # plt.plot(idxs, sorted_all[idxs]*np.max(motif_plot), "g")
    # plt.plot(sorted_all*np.max(motif_plot)/np.max(sorted_all))
    a=np.count_nonzero(motif_mismatches==0)
    b=np.count_nonzero(all_mismatches==0)
    print("#PRE:",motif_ps.size, all_ps.size, motif_ps.size/all_ps.size)
    print("#POST:", a, b, a/float(b))
    # print("#POST:", (motif_ps>=threshold).sum(), (all_ps>=threshold).sum(), (motif_ps>=threshold).sum()/float((all_ps>=threshold).sum()))
    return (a, b)

def count_diplo(results):
    return sum(check_is_diplo(result) for result in results)


def parse_results(filename):
    return [(line.split("\t")[0], eval(line.split("\t")[1]))
            for line in open(filename) if not line.startswith("#")]

def run_comparison(folder, unique=True):
    print(folder)
    if unique:
        macs_results = {i: parse_results("%s%s_macs_unique_summits_alignments_diplotypes.tsv" % (folder, i)) for i in range(1, 6)}
        is_unique = lambda peak, i: peak in graph_unique[str(i)]
    else:
        macs_results = {i: parse_results("%s%s_macs_all_summits_alignments_diplotypes.tsv" % (folder, i)) for i in range(1, 6)}        
        is_unique= lambda peak, i: True
    graph_results = {i: parse_results("%s%s_limited_summits_alignments_diplotypes.tsv" % (folder, i)) for i in range(1, 6)}
    macs_motifs = {line.split("\t")[0]: {peak.strip() for peak in line.split("\t")[1].split(",")} for line in open(folder+"/fimo_macsmotif_summary.tsv")}
    graph_motifs = {line.split("\t")[0]: {peak.strip() for peak in line.split("\t")[1].split(",")} for line in open(folder+"/fimo_graphmotif_summary.tsv")}
    graph_unique = {line.split("\t")[0]: {peak.strip() for peak in line.split("\t")[1].split(",")} for line in open(folder+"/unqiue_summary.tsv")}

    macs_non_motif_results = [result for i, results in macs_results.items()
                              for peak, result in results if  peak not in macs_motifs[str(i)]]
    macs_motif_results = [result for i, results in macs_results.items()
                     for peak, result in results if peak in macs_motifs[str(i)]]
    graph_non_motif_results = [result for i, results in graph_results.items()
                         for peak, result in results if is_unique(peak, i) and peak not in graph_motifs[str(i)]]
    graph_motif_results = [result for i, results in graph_results.items()
                           for peak, result in results if is_unique(peak, i) and peak in graph_motifs[str(i)]]
    # compare_two(graph_motif_results, graph_motif_results+graph_non_motif_results,
    # macs_motif_results, macs_motif_results+macs_non_motif_results)
    # plt.show()
    return graph_motif_results, graph_non_motif_results, macs_motif_results, macs_non_motif_results
    motif_models.simple_model(graph_motif_results, graph_non_motif_results,
                              macs_motif_results, macs_non_motif_results)


def compare_folders(folders):
    results = [[], [], [], []]
    counts = []
    tots = []
    for folder in folders:
        res = run_comparison(folder)

        for i in range(4):
            results[i].extend(res[i])
        print([len(r) for r in res])
        print(folder)
        counts.append([count_diplo(r) for r in res])
        tots.append([len(r) for r in res])
        # compare_two(res[0], res[0]+res[1], res[2], res[2]+res[3])
        # plt.show()
    print(counts)
    print(tots)
    tots = np.array(tots)
    counts = np.array(counts)
    inv = tots-counts
    props1 = counts[:, 0]/(counts[:, 0]+counts[:, 1])
    props2 = counts[:, 2]/(counts[:, 2]+counts[:, 3])
    props3 = inv[:, 0]/(inv[:, 0]+inv[:, 1])
    props4 = inv[:, 2]/(inv[:, 2]+inv[:, 3])
    print("graph1", props1)
    print("graph2", props3)
    print("graphdiff", props1-props3)
    print("macs1", props2)
    print("macs2", props4)
    print("macsdiff", props2-props4)

    # props = counts/tots
    # print(tots.T)
    # print(counts.T)
    # print(props.T)
    res = results
    compare_two(res[0], res[0]+res[1], res[2], res[2]+res[3])
    plt.show()
    motif_models.simple_model(*results)
        

def run(folder, includes="all"):
    #interval_name = "macs_unique_graph_"
    interval_name= ""
    rs = {i: parse_results("%s%s_%slimited_summits_alignments_diplotypes.tsv" % (folder, i, interval_name)) for i in range(1, 6)}
    with_macs = includes == "macs"
    if includes == "macs":
        summary_name = "motif_summary_graph_matching_macs.tsv"
        includes = "shared"
        macs_motifs = {line.split("\t")[0]: {peak.strip() for peak in line.split("\t")[1].split(",")} for line in open(folder+"/" + summary_name)}        
    summary_name = "fimo_graphmotif_summary.tsv"
    motifs = {line.split("\t")[0]: {peak.strip() for peak in line.split("\t")[1].split(",")} for line in open(folder+"/" + summary_name)}        
    print(sum(len(m) for m in motifs.values()))
    # summary_name = "fimo_macs_uniquemotif_summary.tsv"

    unique = {line.split("\t")[0]: {peak.strip() for peak in line.split("\t")[1].split(",")} for line in open(folder+"/unqiue_summary.tsv")}

    # for chrom, motif in motifs.items():
    #     in_chrom = {result[0] for result in rs[int(chrom)]}
    #     missing = [m for m in motif if m not in in_chrom]
    #     assert not missing, missing
    #     # assert all(m in in_chrom for m in motif)
    # 
    # for chrom, _unique in unique.items():
    #     in_chrom = {result[0] for result in rs[int(chrom)]}
    #     assert all(u in in_chrom for u in _unique)

    filter_funcs = {"all": lambda peak, i: True,
                    "unique": lambda peak, i: peak in unique[str(i)],
                    "shared": lambda peak, i: peak not in unique[str(i)]}
    
    non_motif_results = [result for i, results in rs.items()
                         for peak, result in results if filter_funcs[includes](peak, i) and peak not in motifs[str(i)]]
    motif_results = [result for i, results in rs.items()
                     for peak, result in results if filter_funcs[includes](peak, i) and peak in motifs[str(i)]]
    all_results = motif_results + non_motif_results
    plt.title(folder)
    print(folder)
    res = analyze_result(motif_results, all_results)
    if with_macs:
        macs_motif_results = [result for i, results in rs.items()
                         for peak, result in results if filter_funcs[includes](peak, i) and peak in macs_motifs[str(i)]]

        analyze_result(macs_motif_results, all_results, "r")
    # plt.show()
    return res


def run_func(folder, interval_name, func):
    includes = "all"
    rs = {i: parse_results("%s%s_%s_diplotypes.tsv" % (folder, i, interval_name)) for i in range(1, 6)}
    summary_name = "fimo_graphmotif_summary.tsv"
    motifs = {line.split("\t")[0]: {peak.strip() for peak in line.split("\t")[1].split(",")} for line in open(folder+"/" + summary_name)}        
    unique = {line.split("\t")[0]: {peak.strip() for peak in line.split("\t")[1].split(",")} for line in open(folder+"/unqiue_summary.tsv")}
    shared_results = [result for i, results in rs.items()
                      for peak, result in results if peak not in unique[str(i)]]

    unique_results = [result for i, results in rs.items()
                     for peak, result in results if peak in unique[str(i)]]
    print(folder)
    print("Unique")
    func(unique_results)
    print("Shared")
    func(shared_results)
    print("All")
    func(unique_results+shared_results)


def count_haplo(results):
    haplo = sum(result.A_count == result.total for result in results)
    t = len(results)
    if t == 0:
        print(haplo, t, "NA")
    else:
        print(haplo, t, haplo/t)

if __name__ == "__main__":
    if sys.argv[1] == "haplo":
        [run_func(folder+"/", sys.argv[2], count_haplo) for folder in sys.argv[3:]]
        exit()
    if sys.argv[1] == "compare":
        compare_folders(sys.argv[2:])
        # [run_comparisons(folder+"/") for folder in sys.argv[2:]]
        exit()



    folders = sys.argv[1:]
    output = [run(folder+"/", includes="unique") for folder in folders]
    s = np.sum(output, axis=0)
    results = np.array(output).T
    print(" ".join(folder.split("/")[1].split("_")[1] for folder in folders))
    for row in results:
        print(" ".join(str(n) for n in row))
    print(" ".join(str(n) for n in results[0]/results[1]))
    print(s[0], s[1], s[0]/s[1])
    plt.show()



import logging
import offsetbasedgraph as obg
from offsetbasedgraph.graphtraverser import GraphTraverserUsingSequence
import pyvg


class LinearRegion(object):
    def __init__(self, chromosome, start, end):
        self.chromosome = chromosome
        self.start = start
        self.end = end

LRC_REGION = LinearRegion("chr19", 54025634, 55084318)
MHC_REGION = LinearRegion("chr6", 28510119, 33480577)


def bed_intervals_to_graph(obg_graph, linear_path_interval,
                           bed_file_name, graph_start_offset):
    f = open(bed_file_name)
    intervals_on_graph = []
    for line in f:
        l = line.split()
        start = int(l[1])
        end = int(l[2])

        start = start - graph_start_offset
        end = end - graph_start_offset
        intervals_on_graph.append(linear_path_interval.get_subinterval(start, end))

    return intervals_on_graph


def fasta_sequence_to_linear_path_through_graph(
        linear_sequence_fasta_file, sequence_retriever, ob_graph, start_node):
    search_sequence = open(linear_sequence_fasta_file).read()
    print("Length of search sequence: %d" % len(search_sequence))
    traverser = GraphTraverserUsingSequence(
        ob_graph, search_sequence, sequence_retriever)
    traverser.search_from_node(start_node)
    linear_path_interval = traverser.get_interval_found()
    return linear_path_interval


def get_linear_paths_in_graph(ob_graph, vg_graph, write_to_file_name=None):
    assert ob_graph is not None
    intervals = {}
    for path in vg_graph.paths:
        obg_interval = path.to_obg(ob_graph=ob_graph)
        obg_interval.name = path.name
        print("Path name: %s" % path.name)
        intervals[obg_interval.name] = obg_interval

    if write_to_file_name is not None:
        logging.info("Writing linear path to %s" % write_to_file_name)
        collection = obg.IntervalCollection(intervals.values())
        collection.to_file(write_to_file_name, text_file=True)

    return intervals


def create_ob_graph_from_vg(vg_json_graph_file_name, ob_graph_file_name="graph.obg"):
    vg_graph = pyvg.Graph.create_from_file(vg_json_graph_file_name)
    ob_graph = vg_graph.get_offset_based_graph()
    ob_graph.to_file(ob_graph_file_name)
    logging.info("Wrote obgraph to %s" % ob_graph_file_name)


def create_linear_path(ob_graph, vg_graph, path_name="ref", write_to_file="linear_path.intervalcollection"):
    assert ob_graph is not None
    linear_paths = get_linear_paths_in_graph(ob_graph, vg_graph, write_to_file_name=write_to_file)
    ref_path = obg.NumpyIndexedInterval.from_interval(linear_paths[path_name])
    return ref_path



import subprocess
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import logging


class MotifMatcher():

    def __init__(self, ranked_fasta_file_name, meme_motif_file_name, run_fimo=True, background_model_file=None):
        self.fasta_file = ranked_fasta_file_name
        self.meme_file = meme_motif_file_name
        self.run_fimo = run_fimo
        self.peaks_matching_motif = set()
        self.sorted_peaks = []
        self.background_model_file = background_model_file

        self.get_sorted_peak_ids()

        self.get_peaks_matching_motif()

    def get_peaks_matching_motif(self):

        out_dir = "fimo_" + self.fasta_file.replace(".fasta", "")

        if self.run_fimo:
            import os
            bgfile = ""
            if self.background_model_file is not None and self.background_model_file != "None":
                bgfile = "--bgfile " + self.background_model_file
                logging.info("Using background model file %s" % self.background_model_file)
            else:
                logging.warning("Not using background model file when running fimo")

            subprocess.check_output(["fimo %s -oc %s %s %s" % (bgfile, out_dir, self.meme_file, self.fasta_file)], shell=True)

        result_file = out_dir + "/fimo.txt"

        for line in open(result_file):
            l = line.split()
            sequence_id = l[2]
            self.peaks_matching_motif.add(sequence_id)

        #print(self.peaks_matching_motif)

    def get_sorted_peak_ids(self):
        for line in open(self.fasta_file):
            if line.startswith(">"):
                id = line.split()[0].replace(">", "")
                self.sorted_peaks.append(id)

    def compute_true_positives(self):
        true_positives = []
        n_matched = 0
        n_checked = 0
        for peak in self.sorted_peaks:
            if peak in self.peaks_matching_motif:
                n_matched += 1
            n_checked += 1
            true_positives.append(n_matched / n_checked)

        return true_positives


def plot_true_positives(peak_file_sets, meme_file_name, plot_title="", save_to_file=None, run_fimo=True, background_model_file=None):
    colors = ['b', 'g']
    i = 0
    font = {'family' : 'DejaVu Sans',
        'weight' : 'normal',
        'size'   : 8}
    import matplotlib
    import numpy as np
    matplotlib.rc('font', **font)

    plt.figure(figsize=(3.0, 3.0), dpi=300)
    axis = plt.gca()
    # Major ticks every 20, minor ticks every 5
    major_ticks = np.linspace(0, 1.0, 5)
    minor_ticks = np.arange(0, 1, 20)

    possible_yticks = np.array(np.linspace(0, 1, 11))
    #ax.set_yticks(minor_ticks, minor=True)

    #ax.set_xticks(major_ticks)
    #ax.set_xticks(minor_ticks, minor=True)

    max_y = 0.0
    for name, fasta_file_name in peak_file_sets:
        logging.info("Background model file is %s" % background_model_file)
        matcher = MotifMatcher(fasta_file_name, meme_file_name, background_model_file=background_model_file)
        true_positives = matcher.compute_true_positives()
        n_matching = len(matcher.peaks_matching_motif)
        print("Fasta: %s" % fasta_file_name)
        n_tot = len(matcher.sorted_peaks)
        print("N tot: %d" % n_tot)
        print("True positives for %s: %d / %d = %.3f" % (name, n_matching, n_tot, n_matching/n_tot))
        plt.plot(true_positives, color=colors[i], label=name, linewidth=1.0)
        max_y = max(max_y, np.max(true_positives[2:]))
        i += 1
    #axis.set_ylim([0.75,1.0])
    axis.set_xlim(0)
    min_y = n_matching / n_tot
    idx = np.where(possible_yticks > min_y)
    yticks = possible_yticks[idx]
    yticks = yticks[np.where(yticks <= max_y)]
    axis.set_yticks(yticks)
    axis.grid(which='both')

    #plt.xlabel("Number of peaks included (of the total set of peaks sorted descending on score)")
    #plt.ylabel("Proportion of peaks enriched for motif")
    plt.title(plot_title)
    plt.legend()
    if save_to_file is not None:
        plt.savefig(save_to_file, dpi=300)
        print("Saved figure to %s" % save_to_file)
    else:
        plt.show()



import logging
import numpy as np
from .mindense import DensePileup
from .sample import get_fragment_pileup
from .control import get_background_track_from_control,\
    get_background_track_from_input, scale_tracks
from .sparsepvalues import PValuesFinder, PToQValuesMapper, QValuesFinder
from .postprocess import HolesCleaner, SparseMaxPaths
from .sparsediffs import SparseValues
import json
import sys


class Configuration:
    def __init__(self):
        self.read_length = None
        self.fragment_length = None
        self.linear_map_name = None
        self.has_control = False
        self.q_values_threshold = 0.05
        self.global_min = None
        self.keep_duplicates = False

    def copy(self):
        o = Configuration()
        o.read_length = self.read_length
        o.fragment_length = self.fragment_length
        o.linar_map_name = self.linear_map_name
        o.has_control = self.has_control
        o.q_values_threshold = self.q_values_threshold
        o.global_min = self.global_min
        o.keep_duplicates = self.keep_duplicates
        return o


class CallPeaks(object):

    def __init__(self, graph, config, reporter, variant_maps=None):
        self.graph = graph
        self.config = config
        assert self.config.fragment_length > self.config.read_length, \
            "Read length %d is larger than fragment size" % (self.config.read_length)
        self.info = config
        self._reporter = reporter
        self.variant_maps = variant_maps

    def run_pre_callpeaks(self, input_reads, control_reads):
        try:
            sample_pileup = get_fragment_pileup(
                self.graph, input_reads, self.config,
                self._reporter)

            background_func = get_background_track_from_control
        except json.decoder.JSONDecodeError as e:
            logging.debug(e)
            logging.critical("Error when parsing json reads. Input file is not valid JSON."
                             "Turn on debugging (--verbose 2) for more debug info.")
            sys.exit(1)

        if not self.config.has_control:
            background_func = get_background_track_from_input
        control_pileup = background_func(self.graph, control_reads,
                                         self.config,
                                         sample_pileup.touched_nodes)
        scale_tracks(sample_pileup, control_pileup,
                     input_reads.n_reads/control_reads.n_reads)

        self._reporter.add("fragment_pileup", sample_pileup)
        self._reporter.add("touched_nodes", sample_pileup.touched_nodes)
        self._reporter.add("background_track", control_pileup)
        self.touched_nodes = sample_pileup.touched_nodes
        self.control_pileup = control_pileup
        self.sample_pileup = sample_pileup

    def get_p_values(self):
        assert self.sample_pileup is not None
        assert self.control_pileup is not None
        self.p_values_pileup = PValuesFinder(
            self.sample_pileup, self.control_pileup).get_p_values_pileup()
        self.p_values_pileup.track_size = self.graph.node_indexes[-1]
        self._reporter.add("pvalues", self.p_values_pileup)
        self.sample_pileup = None
        self.control_pileup = None

    def get_p_to_q_values_mapping(self):
        assert self.p_values_pileup is not None
        finder = PToQValuesMapper.from_p_values_pileup(
            self.p_values_pileup)
        self.p_to_q_values_mapping = finder.get_p_to_q_values()

    def get_q_values(self):
        assert self.p_values_pileup is not None
        assert self.p_to_q_values_mapping is not None
        finder = QValuesFinder(
            self.p_values_pileup,
            self.p_to_q_values_mapping)

        self.q_values_pileup = finder.get_q_values()
        self.q_values_pileup.track_size = self.p_values_pileup.track_size
        self._reporter.add("qvalues", self.q_values_pileup)

    def call_peaks_from_q_values(self, linear_path=None):
        assert self.q_values_pileup is not None
        caller = CallPeaksFromQvalues(
            self.graph, self.q_values_pileup,
            self.config, self._reporter,
            touched_nodes=self.touched_nodes,
            config=self.config,
            linear_path=linear_path,
            variant_maps=self.variant_maps
            )
        caller.callpeaks()
        self.max_path_peaks = caller.max_paths

    def run(self, input_intervals, control_intervals):
        self.run_to_p_values(input_intervals, control_intervals)
        self.get_p_to_q_values_mapping()
        self.get_q_values()
        self.call_peaks_from_q_values()

    def run_to_p_values(self, input_intervals, control_intervals):
        self.run_pre_callpeaks(input_intervals, control_intervals)
        self.get_p_values()


class CallPeaksFromQvalues:
    def __init__(self, graph, q_values_pileup,
                 experiment_info, reporter,
                 cutoff=0.1, raw_pileup=None, touched_nodes=None,
                 config=None, q_values_max_path=False, linear_path=None, variant_maps=None):

        self.graph = graph
        self.q_values = q_values_pileup
        self.info = experiment_info
        self.cutoff = cutoff
        self.raw_pileup = raw_pileup
        self.touched_nodes = touched_nodes
        self.save_tmp_results_to_file = False
        self.graph_is_partially_ordered = False
        self.q_values_max_path = q_values_max_path
        self.linear_path = linear_path
        self.variant_maps = variant_maps

        if config is not None:
            self.cutoff = config.q_values_threshold
        self._reporter = reporter
        # self.info.to_file(self.out_file_base_name + "experiment_info.pickle")
        logging.info("Using q value cutoff %.4f" % self.cutoff)

    def __threshold(self):
        threshold = -np.log10(self.cutoff)
        logging.info("Thresholding peaks on q value %.4f" % threshold)
        self.pre_processed_peaks = self.q_values.threshold_copy(threshold)
        self._reporter.add("thresholded", self.pre_processed_peaks)

    def __postprocess(self):
        logging.info("Filling small Holes")
        self.pre_processed_peaks = HolesCleaner(
            self.graph,
            self.pre_processed_peaks,
            self.info.read_length,
            self.touched_nodes
        ).run()
        self._reporter.add("hole_cleaned", self.pre_processed_peaks)
        self.filtered_peaks = self.pre_processed_peaks

    def __get_max_paths(self):
        logging.info("Getting maxpaths")
        if not self.q_values_max_path:
            file_name = self._reporter._base_name+"direct_pileup"
            logging.info("Reading raw direct pileup from file %s" % file_name)
            _pileup = SparseValues.from_sparse_files(file_name)
            self.raw_pileup = _pileup
        else:
            # raise NotImplementedError()
            _pileup = self.q_values

        assert(self.graph.uses_numpy_backend)
        logging.info("Running Sparse Max Paths")
        max_paths, sub_graphs = SparseMaxPaths(
            self.filtered_peaks, self.graph, _pileup, self.variant_maps).run()

        self._reporter.add("all_max_paths", max_paths)
        logging.info("All max paths found")

        self.q_values = DensePileup(
            self.graph, self.q_values.to_dense_pileup(
                self.graph.node_indexes[-1]))

        for max_path in max_paths:
            assert max_path.length() >= 0, "Max path %s has negative length" % max_path
            if max_path.length() == 0:
                logging.warning("Max path has 0 length: %s" % max_path)
                max_path.set_score(0)
                continue

            score = np.max(self.q_values.get_interval_values(max_path))
            max_path.set_score(score)
            max_path.chromosome = self._reporter._base_name.replace("_", "")
            assert not np.isnan(score), "Score %s is nan" % score



        pairs = list(zip(max_paths, sub_graphs))
        pairs.sort(key=lambda p: p[0].score, reverse=True)
        logging.info("N unfiltered peaks: %s", len(max_paths))
        pairs = [p for p in pairs if
                 p[0].length() >= self.info.fragment_length]
        logging.info("N filtered peaks: %s", len(pairs))
        self._reporter.add("sub_graphs", [pair[1] for pair in pairs])
        self.max_paths = [p[0] for p in pairs]
        self._reporter.add("max_paths", self.max_paths)

    def callpeaks(self):
        logging.info("Calling peaks")
        self.__threshold()
        self.__postprocess()
        self.__get_max_paths()




import offsetbasedgraph as obg
import numpy as np


class GenoType:
    def __init__(self, start_pos, end_pos, variant_ids):
        self.start_pos = start_pos
        self.end_pos = end_pos
        self.variant_ids = variant_ids


class VariantList:
    def __init__(self, indexed_interval, graph, seq_graph):
        self._indexed_interval = indexed_interval
        self._graph = graph
        self._seq_graph = seq_graph
        self.alts = {}
        self.insertions = {}

    def add_simple_var(self, pos, alt_seq, ref_seq, i):
        for j, pair in enumerate(zip(alt_seq, ref_seq)):
            if pair[0] != pair[1]:
                break
        print(ref_seq, alt_seq)
        alt_seq = alt_seq[j:]
        ref_seq = ref_seq[j:]
        if not alt_seq:
            alt_seq = "."
            
        print(ref_seq, alt_seq)
        pos += j
        assert self._indexed_interval.get_node_offset_at_offset(pos) == 0, (self._indexed_interval.get_node_offset_at_offset(pos), i, j)
        prev_node = self._indexed_interval.get_node_at_offset(pos-1)
        if alt_seq == ".":
            self.insertions[(prev_node, self._indexed_interval.get_node_at_offset(pos+len(ref_seq)))] = i
        else:
            next_nodes = [node for node in self._graph.adj_list[prev_node]
                          if self._seq_graph.get_sequence_on_directed_node(node).lower().startswith(alt_seq.lower())]
            assert len(next_nodes) == 1, (prev_node, next_nodes, i)
            if not next_nodes:
                # print(seq)
                # print([self._seq_graph.get_sequence_on_directed_node(node).lower() for node in self._graph.adj_list[prev_node]])
                return
            next_node = next_nodes[0]
            self.alts[next_node] = i

    def add_var(self, pos, alt_seq, ref_seq, i):
        for seq in alt_seq.split(","):
            self.add_simple_var(pos, seq, ref_seq, i)

    def create_lookup(self, vcf_file_name):
        for i, line in enumerate(open(vcf_file_name)):
            parts = line.split("\t")
            pos = int(parts[1])-1  # to 0-indx
            ref_seq = parts[3]
            alt_seq = parts[4]
            self.add_var(pos, alt_seq, ref_seq, i)

    def create_type(self, interval):
        node_ids = interval.region_paths
        alts = [self.alts[node_id] for node_id in node_ids if node_id in self.alts]
        insertions = [self.insertions[(n1, n2)] for n1, n2 in zip(node_ids[:-1], node_ids[1:])
                      if (n1, n2) in self.insertions]
        start_pos = self._indexed_interval.get_offset_at_position(interval.start_position)
        end_pos = self._indexed_interval.get_offset_at_position(interval.end_position)
        return GenoType(start_pos, end_pos, alts+insertions)

    @classmethod
    def from_name(cls, name):
        indexed_interval = obg.NumpyIndexedInterval.from_file(
            name+"_linear_pathv2.interval")
        graph = obg.Graph.from_file(name+".nobg")
        seq_graph = obg.SequenceGraph.from_file(name+".nobg.sequences")
        var_list = cls(indexed_interval, graph, seq_graph)
        var_list.create_lookup(name+"_variants.vcf")
        return var_list


class GenotypeMatrix:
    def __init__(self, matrix, positions):
        self._matrix = matrix
        self._positions = positions

    def find_path(self, genotype):
        variants = set(genotype.variants)
        start_idx = np.searchsorted(self._positions, genotype.start_pos)
        end_idx = np.searchsorted(self._positions, genotype.end_pos)
        print(start_idx, end_idx)
        relevant_matrix = self._matrix[start_idx:end_idx]
        relevant_ids = range(start_idx, end_idx)
        hits = np.array([variant_id in variants
                         for variant_id in relevant_ids])[:, None]
        return np.flatnonzero(np.all(relevant_matrix == hits, axis=0))

    @classmethod
    def from_vcf(cls, vcf_file_name):
        matrix = []
        positions = []
        for line in open(vcf_file_name):
            if line.startswith("#"):
                continue
            row = line.split("\t")
            positions.append(int(row[1]))
            vals = ["1" in val.split(":")[0] for val in vals[9:]]
            matrix.append(vals)
        return cls(np.array(matrix, dtype="bool"),
                   np.array(positions, dtype="int"))


def main(name, interval_file):
    var_list = VariantList.from_name(name)
    ineterval_collection = IntervalCollection.from_file(interval_file)
    types = [var_list.create_type(interval) for
             interval in interval_collection]
    matrix = GenoTypeMatrix.from_vcf("chr"+name+"_variants.vcf")
    hits_list = [matrix.find_path(genotype) for genotype in types]
    print(hits_list)



from collections import defaultdict
import matplotlib.pyplot as plt
import pyvg
from pyvg.sequences import SequenceRetriever
import offsetbasedgraph as obg
from offsetbasedgraph.graphtraverser import GraphTraverserUsingSequence
from graph_peak_caller.peakcollection import PeakCollection
from offsetbasedgraph import IntervalCollection, DirectedInterval
# from graph_peak_caller.subgraphcollection import SubgraphCollection
from .util import LinearRegion, get_linear_paths_in_graph
from graph_peak_caller.legacy.sparsepileup import SparsePileup
import numpy as np

MHC_REGION = LinearRegion("chr6", 28510119, 33480577)


def create_linear_path(ob_graph, vg_graph):
    assert ob_graph is not None
    linear_paths = get_linear_paths_in_graph(ob_graph, vg_graph, "linear_maps")
    ref_path = linear_paths["ref"].to_indexed_interval()
    return ref_path


def create_linear_peaks_from_bed(linear_sequence_fasta_file, peaks_bed_file,
                                 obg_graph_file_name, vg_graph_file_name,
                                 start_node,
                                 region):

    ob_graph = obg.GraphWithReversals.from_file(obg_graph_file_name)
    search_sequence = open(linear_sequence_fasta_file).read()
    sequence_retriever = SequenceRetriever.from_vg_graph(vg_graph_file_name)
    traverser = GraphTraverserUsingSequence(
        ob_graph, search_sequence, sequence_retriever)
    traverser.search_from_node(start_node)
    linear_path_interval = traverser.get_interval_found()
    IntervalCollection([linear_path_interval]).to_file("linear_path.intervalcollection",
                                                       text_file=True)
    print("Length")
    print(linear_path_interval.length())
    print(linear_path_interval.region_paths[0])
    print(linear_path_interval.start_position)
    print(linear_path_interval.end_position)

    linear_peaks = PeakCollection.create_from_linear_intervals_in_bed_file(
                        obg_graph_file_name,
                        linear_path_interval,
                        peaks_bed_file,
                        region.start,
                        region.end)

    linear_peaks.to_file("linear_peaks.intervalcollection", text_file=True)


class SubgraphAnalyser(object):

    def __init__(self, graph, subgraphs_file_name):
        self.graph = graph
        self.subgraphs = SubgraphCollection.from_pickle(
            subgraphs_file_name, graph=graph)

    def print_sizes(self):
        sizes = []
        for subgraph in self.subgraphs:
            print(subgraph.n_basepairs())
            sizes.append(subgraph.n_basepairs())

        plt.hist(sizes, bins=100)
        plt.show()


class SubgraphComparer():
    # Compare subgraph vs linear peaks
    def __init__(self, graph, subgraphs_file_name, peaks_file_name):
        self.graph = graph
        IntervalCollection.interval_class = DirectedInterval
        self.subgraphs = SubgraphCollection.from_pickle(
            subgraphs_file_name, graph=graph)
        self.peaks = PeakCollection.create_list_from_file(
            peaks_file_name, graph=graph)

    def check_peaks_in_subgraphs(self):
        n_in_subgraphs = 0
        for peak in self.peaks:
            print(peak)
            if self.subgraphs.contains_interval(peak):
                n_in_subgraphs += 1

        print("%d peaks are in subgraphs" % n_in_subgraphs)


def find_missing_graph_peaks():
    ob_graph = obg.GraphWithReversals.from_file("obgraph")
    vg_graph = pyvg.vg.Graph.create_from_file("haplo1kg50-mhc.json")
    path = create_linear_path(ob_graph, vg_graph)
    comparer = PeaksComparer.create_from_graph_peaks_and_linear_peaks(
        "ctcf05_peaks.narrowPeak",
        "laststepmax_paths.intervalcollection",
        ob_graph,
        path,
        MHC_REGION)
    comparer.check_similarity()


def analyse_without_control():
    ob_graph = obg.GraphWithReversals.from_file("haplo1kg50-mhc.obg")
    vg_graph = pyvg.vg.Graph.create_from_file("haplo1kg50-mhc.json")
    path = create_linear_path(ob_graph, vg_graph)
    comparer = PeaksComparer.create_from_graph_peaks_and_linear_peaks(
        "macs_without_control_peaks.narrowPeak",
        #"ctcf_q50_without_control_max_paths.intervalcollection",
        "ctcf_r1_max_paths.intervalcollection",
        ob_graph,
        path,
        MHC_REGION)
    comparer.check_similarity(analyse_first_n_peaks=50)


def get_mappings_overlapping_with_interval(mappings, intervals):
    i = 0
    n = defaultdict(int)
    for i, mapping in enumerate(mappings):

        for j, interval in enumerate(intervals):
            if interval.contains_position(mapping.start_position):
                n[j] += 1

    return n

def analyse_pileups_on_peaks(ob_graph, pileups_file_names, peak_intervals_file_name):
    print("Analysing peaks")
    pileups = {name: SparsePileup.from_bed_graph(ob_graph, pileup) for name, pileup in pileups_file_names.items()}
    peaks = IntervalCollection.from_file(peak_intervals_file_name, text_file=True)

    for peak in peaks:
        print()
        print("Peak %s" % peak)
        rp = peak.region_paths[0]
        for name, pileup in pileups.items():
            pileup_sum = sum(pileup.data[rp].sum() for rp in peak.region_paths)
            print("Pileup %s: %d" % (name, pileup_sum))


def vg_alignments_to_linear():
    ob_graph = obg.GraphWithReversals.from_file("haplo1kg50-mhc.obg")
    vg_graph = pyvg.vg.Graph.create_from_file("haplo1kg50-mhc.json")
    path = create_linear_path(ob_graph, vg_graph)
    analyser = AlignmentsAnalyser(vg_graph, "ENCFF001HNI_haplo1kg50-mhc_filtered_q50.gam", ob_graph, path)  # sample reads
    #linear = analyser.to_linear_alignments()
    #collection = IntervalCollection(linear)
    #collection.to_file("graph_reads_on_linear2.intervals")

    linear = IntervalCollection.from_file("graph_reads_on_linear2.intervals").intervals
    #linear = IntervalCollection.create_list_from_file("graph_reads_on_linear.intervals")
    f = open("graph_reads_on_linear.bed", "w")
    path = path.to_indexed_interval()
    linear_reads = []
    for read in linear:
        read.graph = ob_graph
        assert np.all(np.array(read.region_paths) > 0) or np.all(np.array(read.region_paths) < 0)

        dir = "+"
        if read.region_paths[0] < 0:
            dir = "-"
            read = read.get_reverse()

        graph_start = read.start_position
        graph_end = read.end_position

        linear_start = MHC_REGION.start + path.get_offset_at_position(graph_start)
        linear_end = MHC_REGION.start + path.get_offset_at_position(graph_end)

        f.writelines("chr6\t%d\t%d\t.\t0\t%s\n" % (linear_start, linear_end, dir))
    f.close()






class GraphNotFoundException(Exception):
    pass

class InvalidVgJsonGamFile(Exception):
    def __init__(self, msg=None):
        if msg is None:
            msg = """Vg alignment file is invalid and could not be parsed.
            When input file with .json-ending is used, Graph Peak Caller assumes a vg
            alignment file in json formatet, created by converting a vg .gam file
            by using vg view -aj.
            """
            super(InvalidVgJsonGamFile, self).__init__(msg)

class InvalidPileupInterval(Exception):
    pass


from statsmodels.discrete.discrete_model import Logit
import numpy as np


def check_is_diplo(result):
    return result.total == result.A_count+result.B_count


def simple_model(motif_results_A, non_results_A, motif_results_B, non_results_B):
    all_results = motif_results_A + non_results_A + motif_results_B + non_results_B
    is_diplo = np.array([check_is_diplo(result) for result in all_results], dtype="int")
    total_gpc = (len(motif_results_A)+len(non_results_A))
    is_gpc = np.zeros(len(all_results), dtype="int")
    is_gpc[:total_gpc] = 1
    motif = np.zeros(len(all_results), dtype="int")
    motif[:len(motif_results_A)] = 1
    motif[total_gpc:total_gpc+len(motif_results_B)] = 1
    X = np.hstack((np.ones_like(is_gpc)[:, None], is_gpc[:, None], is_diplo[:, None], (is_gpc*is_diplo)[:, None]))
    print(np.sum(X*motif[:, None], axis=0)/np.sum(X, axis=0))
    print(np.sum((1-X)*motif[:, None], axis=0)/np.sum((1-X), axis=0))
    y = motif
    model = Logit(y, X)
    result = model.fit()
    print(result.summary())



import logging


class PeakFasta:
    def __init__(self, sequence_retriever):
        self._sequence_retriever = sequence_retriever

    def write_max_path_sequences(self, file_name, max_paths):
        f = open(file_name, "w")
        i = 0
        for max_path in max_paths:
            seq = self._sequence_retriever.get_interval_sequence(max_path)
            f.write(">peak" + str(i) + " " +
                    max_path.to_file_line() + "\n" + seq + "\n")
            i += 1
        f.close()
        logging.info("Wrote max path sequences to fasta file: %s" %
                     (file_name))

    def save_intervals(self, out_fasta_file_name, interval_collection):
        f = open(out_fasta_file_name, "w")
        i = 0
        for max_path in interval_collection.intervals:
            seq = self._sequence_retriever.get_interval_sequence(max_path)
            f.write(">peak" + str(i) + " " +
                    max_path.to_file_line() + "\n" + seq + "\n")
            i += 1
            if i % 10000 == 0:
                logging.info("Writing sequence # %d" % i)
        f.close()



import numpy as np


class SimpleValuedIndexes():
    def __init__(self, indexes, values):
        self.indexes = indexes
        self.values = values

    def sum(self):
        lengths = np.diff(self.indexes)
        return np.sum(lengths*self.values)


class RpScore:
    def __init__(self, max_score, sum_score):
        self.sum_score = sum_score
        self.max_score = max_score

    def __getitem__(self, item):
        if item == 0:
            return self.max_score
        elif item == 1:
            return self.sum_score
        else:
            raise NotImplementedError()

    def sum(self):
        return self.sum_score

    def max_value(self):
        return self.max_score

    @classmethod
    def from_valued_indexes(cls, vi):
        return cls(np.max(vi.all_values()), vi.sum())




from collections import defaultdict


class FimoEntry:
    def __init__(self, peak_id, start, stop, direction, score):
        self.peak_id = peak_id
        self._start = start
        self._end = stop
        self._direction = direction
        self._score = score

    @classmethod
    def from_file_line(cls, line):
        parts = line.split()
        peak_id = parts[2]
        start = int(parts[3])
        stop = int(parts[4])
        direction = parts[5]
        score = float(parts[6])
        return cls(peak_id, start, stop, direction, score)

    def __repr__(self):
        return "FE(%s: %s-%s [%s])" % (
            self.peak_id, self._start, self._end, self._direction)


class FimoFile:
    def __init__(self, entry_dict):
        self._entry_dict = entry_dict

    def get_entries(self, peak_id):
        return self._entry_dict[peak_id]

    def get_best_entries(self):
        return [max(entries, key=lambda x: x._score) for entries in
                self._entry_dict.values()]

    @classmethod
    def from_file(cls, filename):
        entries = defaultdict(list)
        for line in open(filename):
            if line.startswith("#"):
                continue
            fimo_entry = FimoEntry.from_file_line(line)
            entries[fimo_entry.peak_id].append(fimo_entry)
        return cls(entries)



class Intervals:
    def __init__(self, intervals):
        self._intervals = intervals
        self.n_reads = 0
        self.n_duplicates = 0

    def _count_and_return(self, x):
        self.n_reads += 1
        return x

    def __iter__(self):
        for interval in self._intervals:
            self.n_reads += 1
            yield interval


class UniqueIntervals:
    def __init__(self, intervals):
        self._intervals = intervals
        self.n_reads = 0
        self.n_duplicates = 0

    def __iter__(self):
        interval_hashes = {}
        for interval in self._intervals:
            _hash = interval.hash(ignore_end_pos=True)
            if _hash in interval_hashes:
                self.n_duplicates += 1
                continue

            interval_hashes[_hash] = True
            self.n_reads += 1
            yield interval



import numpy as np
import logging
from itertools import chain
from collections import defaultdict
from ..sparsediffs import SparseDiffs
from ..custom_exceptions import InvalidPileupInterval


class NodeInfo:
    def __init__(self, dist_dict=None):
        self._dist_dict = dist_dict if dist_dict\
                          is not None else defaultdict(int)

    def update(self, starts):
        for read_id, val in starts.items():
            self._dist_dict[read_id] = max(self._dist_dict[read_id], val)

    def __str__(self):
        return str(self._dist_dict)

    def __repr__(self):
        return str(self._dist_dict)


class SparseGraphPileup:
    def __init__(self, graph):
        self.starts = []
        self.ends = []
        self.node_starts = np.zeros(len(graph.blocks)+2)
        self.touched_nodes = np.zeros(self.node_starts.size, dtype="bool")

    def __str__(self):
        return("\n".join([str(self.starts), str(self.ends),
                          str(self.node_starts)]))


class ReadsAdder:
    def __init__(self, graph, pileup, keep_ends=False):
        self._graph = graph
        self.min_id = graph.min_node
        self._node_indexes = graph.node_indexes
        self._pileup = pileup
        self._pos_ends = {node_id: [] for node_id in graph.blocks.keys()}
        self._neg_ends = {-node_id: [] for node_id
                          in graph.blocks.keys()}

    def _handle_interval(self, interval):
        # self._pileup.add_interval(interval)
        end_pos = interval.end_position
        rp = end_pos.region_path_id
        try:
            if rp < 0:
                self.add_open_neg_interval(interval)
                self._neg_ends[rp].append(end_pos.offset)
            else:
                self.add_open_pos_interval(interval)
                self._pos_ends[rp].append(end_pos.offset)
        except KeyError:
            logging.critical("Error with read/interval %s" % interval)
            print(self._graph)
            raise

    def get_pos_ends(self):
        return self._pos_ends

    def get_neg_ends(self):
        return self._neg_ends

    def _add_start_pos(self, pos):
        rp = (abs(pos.region_path_id)-self.min_id)
        if pos.region_path_id > 0:
            self._pileup.starts.append(self._node_indexes[rp]+pos.offset)
        else:
            self._pileup.ends.append(self._node_indexes[rp+1]-pos.offset)

    def add_open_pos_interval(self, interval):
        rps = [rp-self.min_id for rp in interval.region_paths]
        try:
            self._pileup.touched_nodes[rps] = True
            self._pileup.node_starts[rps[1:]] += 1
            self._pileup.node_starts[1:][rps[:-1]] -= 1
            self._add_start_pos(interval.start_position)
        except IndexError as e:
            raise InvalidPileupInterval("Interval has node(s) not part of "
                                        "graph/pileup: %s. Message %s" % (interval.region_paths, e))

    def add_open_neg_interval(self, interval):
        rps = [abs(rp)-self.min_id for rp in interval.region_paths]
        self._pileup.touched_nodes[rps] = True
        self._pileup.node_starts[1:][rps[1:]] -= 1
        self._pileup.node_starts[rps[:-1]] += 1
        self._add_start_pos(interval.start_position)

    def add_reads(self, intervals):
        i = 0
        for interval in intervals:
            if i % 5000 == 0:
                logging.info("%d reads processed" % i)
            self._handle_interval(interval)
            i += 1


class ReadsAdderWDirect(ReadsAdder):
    def __init__(self, graph, pileup):
        self.pos_read_ends = []
        self.neg_read_ends = []
        super().__init__(graph, pileup)

    def _handle_interval(self, interval):
        # self._pileup.add_interval(interval)
        super()._handle_interval(interval)
        end_pos = interval.end_position
        rp = end_pos.region_path_id
        if rp < 0:
            self.neg_read_ends.append(
                self._node_indexes[abs(rp)+1-self.min_id]-end_pos.offset)
        else:
            self.pos_read_ends.append(
                self._node_indexes[abs(rp)-self.min_id]+end_pos.offset)


class SparseExtender:
    def __init__(self, graph, pileup, fragment_length):
        self._graph = graph
        self._pileup = pileup
        self._fragment_length = fragment_length
        if self._fragment_length <= 0:
            raise Exception("Invalid fragment length %d used in SparseExtender" % self._fragment_length)

        self._set_adj_list()
        self._graph_size = graph.node_indexes[-1]

    def _set_adj_list(self):
        self._adj_list = self._graph.adj_list

    def get_node_ids(self):
        return self._graph.get_topological_sorted_node_ids()

    def _add_node_end(self, node_id, value):
        self._pileup.node_starts[node_id+1-self._graph.min_node] -= value

    def _add_node_start(self, node_id, value):
        self._pileup.node_starts[node_id-self._graph.min_node] += value

    def _add_end(self, index):
        self._pileup.ends.append(index)

    def _get_array_idx(self, node_id):
        return self._graph.node_indexes[node_id-self._graph.min_node]

    def run_linear(self, starts_dict):
        node_ids = self.get_node_ids()
        node_infos = defaultdict(NodeInfo)
        cur_id = 0
        empty = NodeInfo()
        cur_array_idx = 0
        n_nodes = len(node_ids)
        for counter, node_id in enumerate(node_ids):
            cur_array_idx = self._get_array_idx(node_id)
            if counter % 1000000 == 0:
                logging.info("Handling node %s of %s", counter, n_nodes)
            info = node_infos.pop(node_id, empty)
            if info._dist_dict:
                self._pileup.touched_nodes[
                    abs(node_id)-self._graph.min_node] = True
            node_size = self._graph.node_size(node_id)
            starts = starts_dict[node_id]
            n_starts = len(starts)
            endsidxs = chain(enumerate(
                (start+self._fragment_length for start in starts),
                start=cur_id),
                             info._dist_dict.items())
            self._add_node_start(node_id, len(info._dist_dict))
            d = {}
            for idx, end in endsidxs:
                if end <= node_size:
                    self._add_end(cur_array_idx+end)
                else:
                    d[idx] = end-node_size
            # cur_array_idx += node_size
            cur_id = cur_id + n_starts
            self._add_node_end(node_id, len(d))
            for next_node in self._adj_list[node_id]:
                node_infos[next_node].update(d)

    def get_pileup(self):
        return np.cumsum(self._pileup)


class ReverseSparseExtender(SparseExtender):
    def _get_array_idx(self, node_id):
        return self._graph.node_indexes[-1] - self._graph.node_indexes[abs(node_id)-self._graph.min_node+1]

    def get_pileup(self):
        return np.cumsum(self._pileup[:-1])[::-1]

    def _set_adj_list(self):
        self._adj_list = self._graph.reverse_adj_list

    def get_node_ids(self):
        return [-node_id for node_id in
                self._graph.get_topological_sorted_node_ids()[::-1]]

    def _add_node_end(self, node_id, value):
        self._pileup.node_starts[-node_id-self._graph.min_node] += value

    def _add_node_start(self, node_id, value):
        self._pileup.node_starts[-node_id+1-self._graph.min_node] -= value

    def _add_end(self, index):
        self._pileup.starts.append(self._graph_size-index)


class SamplePileupGenerator:
    def __init__(self, graph, extension):
        logging.info("Using extension %d when extending reads. " % extension)
        if extension < 0:
            raise Exception("Invalid extension size %d used. Must be positive. Is fragment length < read length?" % extension)
        self._pileup = SparseGraphPileup(graph)
        self._graph = graph
        self._reads_adder = ReadsAdderWDirect(graph, self._pileup)
        self._pos_extender = SparseExtender(graph, self._pileup, extension)
        self._neg_extender = ReverseSparseExtender(
            graph, self._pileup, extension)

    def get_direct_pileup(self):
        new_pileup = SparseGraphPileup(self._graph)
        new_pileup.ends = self._pileup.ends + self._reads_adder.pos_read_ends
        new_pileup.starts = self._pileup.starts + self._reads_adder.neg_read_ends
        new_pileup.node_starts = self._pileup.node_starts

        sparsediff = SparseDiffs.from_pileup(
            new_pileup, self._graph.node_indexes)
        sparsediff.clean()
        sparse_values = sparsediff.get_sparse_values()
        sparse_values.track_size = self._graph.node_indexes[-1]
        return sparse_values

    def run(self, reads, reporter=None):
        self._reads_adder.add_reads(reads)
        if reporter is not None:
            reporter.add("direct_pileup", self.get_direct_pileup())
        self._pos_extender.run_linear(self._reads_adder.get_pos_ends())
        self._neg_extender.run_linear(self._reads_adder.get_neg_ends())
        sdiffs = SparseDiffs.from_pileup(self._pileup,
                                         self._graph.node_indexes)
        sdiffs.touched_nodes = set(
            np.flatnonzero(
                self._pileup.touched_nodes[:-2]) + self._graph.min_node)
        return sdiffs



import numpy as np
from offsetbasedgraph.vcfmap import next_node_func, prev_node_func


def get_end_values_func(pileup, graph):
    node_indexes = graph.node_indexes

    def get_end_values(node_ids):
        node_idxs = node_indexes[node_ids+1]-1
        pileup_idxs = np.digitize(node_idxs, pileup.indices)-1
        return pileup.values[pileup_idxs]

    return get_end_values


def get_start_values_func(pileup, graph):
    node_indexes = graph.node_indexes

    def get_start_values(node_ids):
        node_idxs = node_indexes[node_ids]
        pileup_idxs = np.digitize(node_idxs, pileup.indices)-1
        return pileup.values[pileup_idxs]

    return get_start_values


def find_invalid_insertions(pileup, insertion_map, graph, reference):
    get_next_ref = next_node_func(graph, reference)
    get_prev_ref = prev_node_func(graph, reference)
    get_end_value = get_end_values_func(pileup, graph)
    get_start_value = get_start_values_func(pileup, graph)


    insertion_nodes = np.flatnonzero(insertion_map) + graph.min_node
    prev_nodes = np.array([get_prev_ref(node)-graph.min_node for node in insertion_nodes])
    next_nodes = np.array([get_next_ref(node)-graph.min_node for node in insertion_nodes])
    last_values = get_end_value(prev_nodes)
    first_values = get_start_value(next_nodes)
    avg_values = (get_start_value(insertion_nodes)+get_end_value(insertion_nodes))/2
    valid_nodes = avg_values*2 > (first_values+last_values)/2
    in_valid_map = insertion_map.copy() > 0
    print("Insertions: %s" % np.count_nonzero(insertion_map))
    in_valid_map[insertion_nodes[valid_nodes]] = 0
    print("InvalidInsertions: %s" % np.count_nonzero(in_valid_map))
    return in_valid_map




import logging

from .controlgenerator import SparseControl
from .linearmap import LinearMap


def get_background_track(graph, intervals, config, extensions,
                         touched_nodes=None):
    sc = SparseControl(config.linear_map_name, graph, extensions,
                       config.fragment_length, touched_nodes)
    if config.global_min is not None:
        sc.set_min_value(config.global_min)
    return sc.create(intervals)


def get_background_track_from_control(
        graph, intervals, config, touched_nodes=None):
    logging.info("Creating background from control reads")
    extensions = [config.fragment_length, 1000, 10000]
    return get_background_track(graph, intervals, config,
                                extensions, touched_nodes)


def get_background_track_from_input(
        graph, intervals, config, touched_nodes=None):
    logging.info("Creating background from input reads")
    extensions = [10000]
    return get_background_track(graph, intervals, config,
                                extensions, touched_nodes)


def scale_tracks(fragment_pileup, background_track, ratio):
    logging.info("Scaling tracks to ratio: %d" % ratio)
    if ratio == 1:
        return

    if ratio > 1:
        logging.warning("More reads in sample than in control")
        fragment_pileup *= 1/ratio
    else:
        logging.info("Scaling control pileup down with ratio %.3f" % ratio)
        background_track *= ratio



from collections import defaultdict
import json
from offsetbasedgraph import graph, Interval, Translation
from Bio import SeqIO
import re


def create_graph_from_vg_json(file_name):
    f = open(file_name)
    block_sequences = {}

    lines = f.readlines()
    line = json.loads(lines[0])
    nodes = line["node"]
    edges = line["edge"]

    import offsetbasedgraph

    blocks_offset_based_graph = {}
    edges_offset_based_graph = defaultdict(list)
    for node in nodes:
        id = node["id"]
        length = len(node["sequence"])
        #print(id, length)
        blocks_offset_based_graph[id] = length
        block_sequences[id] = node["sequence"].lower()

    for edge in edges:
        edges_offset_based_graph[edge["from"]].append(edge["to"])

    g = graph.Graph(blocks_offset_based_graph, edges_offset_based_graph)

    return g, block_sequences

def create_translation_from_vg_json(json_file_name, g, block_sequences):
    chromosome, chromosome_sequence = map_subgraph_to_chromosome("dm6.fa", g, block_sequences)
    path = find_block_through_graph_matching_sequence(chromosome_sequence, g, block_sequences)

    linear_blocks = {}
    linear_blocks[chromosome] = len(chromosome_sequence)

    linear_graph = graph.Graph(linear_blocks, {})
    last_block = path[-1]
    last_block_length = len(block_sequences[last_block])
    path_interval = Interval(0, last_block_length, path)
    forward_trans = {}
    forward_trans[chromosome] = [path_interval]

    backward_trans = {}
    offset = 0
    for block in path:
        block_length = len(block_sequences[block])
        backward_trans[block] = [Interval(0, block_length, [chromosome], linear_graph)]
        offset += block_length

    trans = Translation(forward_trans, backward_trans, linear_graph)

    return trans


def map_subgraph_to_chromosome(fasta_file, sub_graph, block_sequences):
    """
    Searches the fasta file for the chromosome
    with starting sequence that matches the beginning of the graph.
    Returns the chromosome name.
    """
    start_blocks = g.get_first_blocks()
    assert len(start_blocks) == 1, "len start blocks != 1"
    start_block = start_blocks[0]
    start_sequence = block_sequences[start_block]

    return find_chromosome_with_sequence(fasta_file, start_sequence)



def find_chromosome_with_sequence(fasta_file, sequence):

    fasta_sequences = SeqIO.parse(open(fasta_file),'fasta')
    for fasta in fasta_sequences:
        name, fasta_sequence = fasta.id, str(fasta.seq)
        print(name)
        if re.match(sequence, fasta_sequence, re.I):  # Case insensitive startswith check
            print("found")
            print(name)
            return name, fasta_sequence

    assert False, "Could not find start sequence %s in fasta file" % sequence
    return False


def find_block_through_graph_matching_sequence(sequence, graph, block_sequences):
    """
    Returns a list of blocks (that are connected) matching the sequence
    """
    sequence = sequence.lower()
    first_block = graph.get_first_blocks()[0]
    path = [first_block]
    current_block = first_block
    offset = len(block_sequences[first_block])
    assert sequence.startswith(block_sequences[first_block]), "%s does not start with %s" % (sequence[0:200], block_sequences[first_block])

    while True:
        edges = graph.adj_list[current_block]
        if len(edges) == 0:
            break

        next_block = None
        for potential_next in edges:
            if potential_next not in graph.blocks:
                return path

            if sequence[offset:].startswith(block_sequences[potential_next]):
                next_block = potential_next
                #print("  Match with %s "  % potential_next)
                break
            #else:
            #    print("  Potential edge %s did not match" % potential_next)

        assert next_block is not None
        #print("Next block: %s" % next_block)

        offset += len(block_sequences[next_block])
        path.append(next_block)
        current_block = next_block

    return path



g, block_sequences = create_graph_from_vg_json("x.json")
trans = create_translation_from_vg_json("x.json", g, block_sequences)

print(trans)







import numpy as np


class LinearIntervalCollection(object):

    def __init__(self, starts, ends):
        self.starts = np.asanyarray(starts)
        self.ends = np.asanyarray(ends)
        self.n_intervals = self.starts.size

    def __eq__(self, other):
        if not np.allclose(self.starts, other.starts):
            return False
        print(self.ends)
        print(other.ends)
        return np.allclose(self.ends, other.ends)

    def __repr__(self):
        s = " ".join(str(i) for i in self.starts)
        e = " ".join(str(i) for i in self.ends)
        return "(%s:%s)" % (s, e)

    def extend_np(self, extension_size):
        return np.add.outer(np.array([-extension_size, extension_size]),
                            self.starts)

    # def n_basepairs_covered(self):
    #     return np.sum(np.abs(self.ends - self.starts))



import logging
from itertools import chain
import scipy.sparse.csgraph as csgraph
import numpy as np
from scipy.sparse import csr_matrix
from .subgraphanalyzer import SubGraphAnalyzer


class DummyTouched:
    def __contains__(self, item):
        return True


class StubsFilter:
    def __init__(self, starts, fulls, ends, graph,
                 last_node=None, touched_nodes=None):
        self._last_node = last_node
        self._graph = graph
        self._starts = starts
        self._fulls = fulls
        self._ends = ends

        self._touched_nodes = touched_nodes
        if touched_nodes is None:
            self._touched_nodes = DummyTouched()

        self._starts_mask = np.ones_like(starts, dtype="bool")
        self._fulls_mask = np.ones_like(fulls, dtype="bool")
        self._ends_mask = np.ones_like(ends, dtype="bool")

        self.filter_start_stubs()
        self.filter_end_stubs()
        self.filtered_ends = self._ends[self._ends_mask]
        self.filtered_starts = self._starts[self._starts_mask]
        self.filtered_fulls = self._fulls[self._fulls_mask]
        self._set_pos_nodes()

        self._full_starts = np.flatnonzero(
            self.find_sub_starts(self.filtered_fulls))
        self._end_starts = np.flatnonzero(
            self.find_sub_starts(self.filtered_ends))
        self._full_ends = np.flatnonzero(
            self.find_sub_ends(self.filtered_fulls))
        self._start_ends = np.flatnonzero(
            self.find_sub_ends(self.filtered_starts))

    def find_sub_starts(self, nodes):
        r = np.array([not all(-adj in self._pos_from_nodes or -adj
                              not in self._touched_nodes
                              for adj in self._graph.reverse_adj_list[-node])
                      for node in nodes], dtype="bool")
        return r

    def find_sub_ends(self, nodes):
        a = np.array([not all(adj in self._pos_to_nodes or adj > self._last_node or adj not in self._touched_nodes
                              for adj in self._graph.adj_list[node])
                      for node in nodes], dtype="bool")
        return a

    def _get_start_filter(self, nodes):
        return np.array([len(self._graph.reverse_adj_list[-node_id])
                         for node_id in nodes], dtype="bool")

    def _get_ends_filter(self, nodes):
        return np.array([len(self._graph.adj_list[node_id])
                         for node_id in nodes], dtype="bool")

    def filter_start_stubs(self):
        """ Locate nodes that are start_nodes of graph"""
        self._fulls_mask &= self._get_start_filter(self._fulls)
        self._ends_mask &= self._get_start_filter(self._ends)

    def filter_end_stubs(self):
        self._starts_mask &= self._get_ends_filter(self._starts)
        self._fulls_mask &= self._get_ends_filter(self._fulls)

    def _set_pos_nodes(self):
        self._pos_to_nodes = set(chain(self._fulls, self._ends))
        self._pos_from_nodes = set(chain(self._starts, self._fulls))


class PosStubFilter(StubsFilter):

    def _set_pos_nodes(self):
        self._pos_to_nodes = set(chain(self._fulls, self._ends))
        self._pos_from_nodes = set(chain(self._starts, self._fulls))

    def find_sub_starts(self, nodes):
        return np.array([not any(-adj in self._pos_from_nodes
                                 for adj in self._graph.reverse_adj_list[-int(node)])
                         for node in nodes], dtype="bool")

    def find_sub_ends(self, nodes):
        return np.array([not any(adj in self._pos_to_nodes
                                 for adj in self._graph.adj_list[int(node)])
                         for node in nodes], dtype="bool")

    def filter_start_stubs(self):
        """ Locate nodes that are start_nodes of graph"""
        pass

    def filter_end_stubs(self):
        pass


class SubGraph:
    def __init__(self, node_ids, graph):
        self._node_ids = node_ids
        self._graph = graph

    def __str__(self):
        return "Subgraph (%s)" % self._node_ids

    def __repr__(self):
        return self.__str__()


class LineGraph:
    stub_class = StubsFilter
    def __init__(self, starts, full, ends, ob_graph, last_node=None, touched_nodes=None):
        if last_node is not None:
            last_node += ob_graph.min_node-1
        starts[0] += ob_graph.min_node-1
        full[0] += ob_graph.min_node-1
        ends[0] += ob_graph.min_node-1
        self.filtered = self.stub_class(starts[0], full[0], ends[0],
                                        ob_graph, last_node, touched_nodes)
        self.full_size = full[1][self.filtered._fulls_mask]
        self.start_size = starts[1][self.filtered._starts_mask]
        self.end_size = ends[1][self.filtered._ends_mask]

        self.kept_starts = np.vstack((starts[0][~self.filtered._starts_mask],
                                      starts[1][~self.filtered._starts_mask]))
        self.kept_ends = np.vstack((ends[0][~self.filtered._ends_mask],
                                    ends[1][~self.filtered._ends_mask]))
        self.kept_fulls = full[0][~self.filtered._fulls_mask]
        self.kept_fulls = self.kept_fulls.astype("int")

        self.full_nodes = self.filtered.filtered_fulls
        self.start_nodes = self.filtered.filtered_starts
        self.end_nodes = self.filtered.filtered_ends
        self._all_nodes = np.r_[self.start_nodes,
                                self.full_nodes,
                                self.end_nodes]
        self._all_nodes = self._all_nodes.astype("int")
        self._all_sizes = np.r_[self.start_size,
                                self.full_size,
                                self.end_size]
        self._all_sizes = self._all_sizes.astype("int")
        self.ob_graph = ob_graph

        self.n_starts = self.start_nodes.size
        self.n_ends = self.end_nodes.size
        self.n_nodes = self._all_nodes.size
        self._matrix = self.make_graph()

    def make_graph(self):
        n_starts = self.start_nodes.size
        n_ends = self.end_nodes.size
        to_nodes_dict = {node: n_starts+i for i, node in
                         enumerate(self._all_nodes[n_starts:])}
        self.end_stub = self._all_nodes.size
        from_nodes = []
        to_nodes = []
        sizes = []
        possible_to_nodes = set(list(self._all_nodes[n_starts:]))
        for i, node_id in enumerate(self._all_nodes[:self._all_nodes.size-n_ends]):
            adj_nodes = self.ob_graph.adj_list[node_id]
            next_nodes = [to_nodes_dict[node] for node in
                          adj_nodes if node in possible_to_nodes]
            from_nodes.extend([i]*len(next_nodes))
            to_nodes.extend(next_nodes)
            sizes.extend([self._all_sizes[i]]*len(next_nodes))

        end_nodes = np.r_[self.filtered._start_ends,
                          n_starts + self.filtered._full_ends,
                          np.arange(self.n_nodes-n_ends, self.n_nodes)]
        to_nodes.extend([self.end_stub]*end_nodes.size)
        from_nodes.extend(list(end_nodes))
        sizes.extend(self._all_sizes[end_nodes])
        # self.n_starts = n_starts
        self._graph_node_sizes = sizes
        return csr_matrix((np.array(sizes, dtype="int32"),
                           (np.array(from_nodes, dtype="int32"),
                            np.array(to_nodes, dtype="int32"))),
                          [self.end_stub+1, self.end_stub+1])

    def get_masked(self, mask):
        n_starts, n_ends = self.n_starts, self.n_ends
        self._all_nodes -= (self.ob_graph.min_node-1)
        T = self._all_nodes.size
        starts = np.vstack((self._all_nodes[:n_starts][mask[:n_starts]],
                            self._all_sizes[:n_starts][mask[:n_starts]]))
        fulls = self._all_nodes[n_starts:T-n_ends][mask[n_starts:T-n_ends]]
        ends = np.vstack((self._all_nodes[T-n_ends:][mask[T-n_ends:]],
                          self._all_sizes[T-n_ends:][mask[T-n_ends:]]))
        d = self.ob_graph.min_node-1
        self.kept_starts[0] -= d
        self.kept_ends[0] -= d
        self.kept_fulls -= d
        starts = np.hstack((starts, self.kept_starts))
        ends = np.hstack((ends, self.kept_ends))
        fulls = np.r_[fulls, self.kept_fulls]

        return starts, fulls, ends

    def filter_small(self, max_size):
        start_nodes = np.r_[np.arange(self.n_starts),
                            self.n_starts+self.filtered._full_starts,
                            self.n_nodes-self.n_ends+self.filtered._end_starts]
        if not start_nodes.size:
            return np.array([], dtype="bool")
        connected_components = csgraph.connected_components(self._matrix)
        shortest_paths = csgraph.shortest_path(self._matrix)
        to_dist = np.min(shortest_paths[start_nodes], axis=0)
        from_dist = shortest_paths[:, self.end_stub]
        return (to_dist+from_dist)[:-1] > max_size


class DividedLinegraph(LineGraph):

    def _get_subgraph(self, idxs):
        self._lookup[idxs] = np.arange(idxs.size)
        sub_matrix = self._matrix[idxs]
        new_indices = self._lookup[sub_matrix.indices]
        sub_matrix = csr_matrix((sub_matrix.data,
                                 new_indices,
                                 sub_matrix.indptr),
                                shape=(idxs.size, idxs.size))
        return sub_matrix

    def filter_small(self, max_size):
        start_nodes = np.r_[np.arange(self.n_starts),
                            self.n_starts+self.filtered._full_starts,
                            self.n_nodes-self.n_ends+self.filtered._end_starts]
        if not start_nodes.size:
            return np.ones_like(self._all_nodes, dtype="bool")
        start_nodes_mask = np.zeros(self.end_stub, dtype="bool")
        start_nodes_mask[start_nodes] = True
        if not start_nodes.size:
            return np.array([], dtype="bool")
        n_components, connected_components = csgraph.connected_components(
            self._matrix[:self.end_stub, :self.end_stub])
        logging.info("Found %s components", n_components)
        complete_mask = np.zeros(self.end_stub, dtype="bool")
        self._lookup = np.empty(self.end_stub+1, dtype="int")
        for comp in range(n_components):
            if comp % 1000 == 0:
                logging.info("Component %s of %s", comp, n_components)

            idxs = np.r_[np.flatnonzero(
                connected_components == comp), self.end_stub]

            if idxs.size-1 > 36:
                complete_mask[idxs[:-1]] = True
                continue
            my_mask = start_nodes_mask[idxs[:-1]]
            start_nodes = np.flatnonzero(my_mask)
            if not start_nodes.size:
                complete_mask[idxs[:-1]] = True
                continue
            subgraph = self._get_subgraph(idxs)
            shortest_paths = csgraph.shortest_path(subgraph)
            to_dist = np.min(shortest_paths[start_nodes], axis=0)
            from_dist = shortest_paths[:, -1]
            complete_mask[idxs[:-1]] = (to_dist+from_dist)[:-1] > max_size
        return complete_mask


class PosDividedLineGraph(DividedLinegraph):
    stub_class = PosStubFilter

    def _backtrace(self, dists, predecessors, start_nodes):
        start = start_nodes[np.argmin(dists[start_nodes, -1])]
        max_row = predecessors[start]
        cur = max_row[-1]
        path = []
        while cur > 0:
            path.append(cur)
            cur = max_row[cur]
        if (not path) or path[-1] != start:
            path += [start]
        return path

    def get_connected_components(self):
        start_nodes = np.r_[np.arange(self.n_starts),
                            self.n_starts+self.filtered._full_starts,
                            self.n_nodes-self.n_ends+self.filtered._end_starts]
        if not start_nodes.size:
            return [], [], []
        start_nodes_mask = np.zeros(self.end_stub, dtype="bool")
        start_nodes_mask[start_nodes] = True
        self._matrix.data += 1
        n_components, connected_components = csgraph.connected_components(
            self._matrix[:self.end_stub, :self.end_stub])
        logging.info("Found %s components", n_components)
        
        # args = np.argsort(connected_components)
        # diffs = np.r_[-1, np.flatnonzero(np.diff(connected_components[args]), args.size-1]
        # component_dict = {i+1: args[diffs[i]+1:diffs[i+1]+1 for i in range(0, len(diffs)-1)}}
        # component_dict[0] = args[:diffs[0]]
        components = []
        subgraphs = []
        for comp in range(n_components):
            idxs = np.flatnonzero(connected_components == comp)
            subgraph = -self._matrix[idxs][:, idxs]
            subgraph.data = subgraph.data + 1
            components.append(idxs)
            subgraphs.append(subgraph)
        return components, subgraphs
    # return {comp: np.flatnonzero(connected_components == comp) for comp in range(n_components)}
    # subgraph = -self._matrix[idxs][:, idxs]
    # subgraph.data = subgraph.data + 1

                

    def max_paths(self):
        start_nodes = np.r_[np.arange(self.n_starts),
                            self.n_starts+self.filtered._full_starts,
                            self.n_nodes-self.n_ends+self.filtered._end_starts]
        if not start_nodes.size:
            return [], [], []
        start_nodes_mask = np.zeros(self.end_stub, dtype="bool")
        start_nodes_mask[start_nodes] = True
        paths = []
        infos = []
        self._matrix.data += 1
        subgraphs = []
        n_components, connected_components = csgraph.connected_components(
            self._matrix[:self.end_stub, :self.end_stub])
        logging.info("Found %s components", n_components)
        # node_id_refs = []
        # subgraphs = []
        for comp in range(n_components):
            if comp % 100 == 0:
                logging.info("Component %s of %s", comp, n_components)
            idxs = np.r_[np.flatnonzero(
                connected_components == comp), self.end_stub]
            subgraph = -self._matrix[idxs][:, idxs]
            subgraph.data = subgraph.data + 1
            subgraphs.append(SubGraph(self._all_nodes[idxs[:-1]], subgraph))
            if idxs.size == 2:
                paths.append([idxs[0]])
                infos.append((0, 0))
                continue
            distances, predecessors = csgraph.shortest_path(
                subgraph, return_predecessors=True, method="BF")
            my_mask = start_nodes_mask[idxs[:-1]]
            start_nodes = np.flatnonzero(my_mask)
            local_idxs = self._backtrace(distances, predecessors, start_nodes)
            global_idxs = idxs[local_idxs]
            paths.append(global_idxs[::-1])
            infos.append(
                SubGraphAnalyzer(subgraph, local_idxs[::-1], distances).get_info())
        return paths, infos, subgraphs




#!/usr/bin/python3
import argparse
import sys
import logging
from graph_peak_caller.logging_config import set_logging_config
from graph_peak_caller.custom_exceptions import *
import pyvg

import matplotlib as mpl
mpl.use('Agg')  # Required for server usage (e.g. travis)

import offsetbasedgraph as obg
import offsetbasedgraph.vcfmap
from graph_peak_caller.peakcollection import Peak, PeakCollection
from graph_peak_caller.sparsediffs import SparseValues
from graph_peak_caller.mindense import DensePileup


from graph_peak_caller.callpeaks_interface import \
    run_callpeaks_interface, run_callpeaks_whole_genome,\
    run_callpeaks_whole_genome_from_p_values, run_callpeaks2, create_alignment_fasta

from graph_peak_caller.analysis.analysis_interface import analyse_peaks_whole_genome,\
    analyse_peaks, peaks_to_fasta, linear_peaks_to_fasta,\
    analyse_manually_classified_peaks, differential_expression,\
    plot_motif_enrichment, get_summits, get_super_summits, peaks_to_linear, move_linear_reads_to_graph,\
    find_linear_path, concatenate_sequence_files, check_haplotype, get_motif_locations,\
    get_haplotype_sequence, get_overlapping_alignments, get_analysis_summaries, \
    split_peaks_by_chromosome

from graph_peak_caller.preprocess_interface import \
    count_unique_reads_interface, create_ob_graph,\
    create_linear_map_interface,\
    split_vg_json_reads_into_chromosomes, shift_estimation


from offsetbasedgraph.interval import NoLinearProjectionException

def main():
    run_argument_parser(sys.argv[1:])


def version(args):
    print("Graph Peak Caller v1.1.1")


def clean_vcf_wrapper(args):
    for chromosome in args.chromosomes.split(","):
        offsetbasedgraph.vcfmap.simplify_vcf(chromosome, args.data_folder)


def project_alignments(alignments, linear_path):
    import numpy as np
    i = 0
    for alignment in alignments:
        if i % 10000 == 0:
            logging.info("Processed %d alignments" % i)
        i += 1

        try:
            if alignment.region_paths[0] < 0:
                assert np.all(np.array(alignment.region_paths) < 0)
                start, end = alignment.get_reverse().to_linear_offsets2(linear_path)
                strand = "-"
            else:
                assert np.all(np.array(alignment.region_paths) > 0)
                start, end = alignment.to_linear_offsets2(linear_path)
                strand = "+"
        except NoLinearProjectionException:
            logging.warning("Found no linear projection for %s. Skipping" % alignment)

        yield start, end, strand

def project_vg_alignments(args):
    from pyvg.conversion import vg_json_file_to_intervals
    linear_path = obg.NumpyIndexedInterval.from_file(args.linear_path_file_name)
    #print(linear_path.nodes_in_interval())
    alignments = vg_json_file_to_intervals(args.alignments_json_file_name, args.graph)

    with open(args.out_file_name, "w") as f:
        projected_alignments = project_alignments(alignments, linear_path)
        for start, end, strand in projected_alignments:
            f.writelines(["%s\t%d\t%d\t.\t0\t%s\n" % (args.chromosome, start, end, strand)])


def check_pruned_graphs_stats(args):
    from pyvg.util import get_stats
    chromosomes = args.chromosomes.split(",")
    graphs_dir = args.graphs_dir

    nodes_pruned = {}
    lengths_pruned = {}
    edges_pruned = {}

    for chromosome in chromosomes:
        stats_pruned = get_stats(graphs_dir + "/" + chromosome + ".prune.vg")
        stats_nonpruned = get_stats(graphs_dir + "/" + chromosome + ".vg")

        print(stats_pruned)
        print(stats_nonpruned)


def vg_json_alignments_to_intervals(args):
    from pyvg.conversion import vg_json_file_to_interval_collection
    interval_collection = vg_json_file_to_interval_collection(args.vg_json_file_name, args.graph)
    interval_collection.to_file(args.out_file_name, text_file=True)
    logging.info("Wrote to file %s" % args.out_file_name)


def get_intersecting_intervals(args):
    from offsetbasedgraph import IntervalCollection
    intervals1 = IntervalCollection.from_file(args.file1, text_file=True, graph=args.graph)
    intervals2 = IntervalCollection.from_file(args.file2, text_file=True, graph=args.graph)

    out = []
    for interval1 in intervals1.intervals:
        for interval2 in intervals2.intervals:
            if interval1.intersects(interval2):
                out.append(interval1)
                logging.info("Found match between %s and %s" % (interval1, interval2))
                continue

    IntervalCollection(out).to_file(args.out_file_name, text_file=True)
    logging.info("Wrote intersecting intervals to %s" % args.out_file_name)



interface = \
{

    'callpeaks':
        {
            'help': 'Call peaks one or multiple graphs.',
            'arguments':
                [
                    ('-g/--graph', 'Graph(s) file name. Either a single file name, e.g. graph.nobg or a '
                                   'pattern with the * wildcard, such as graph_*.nobg. In the latter case'
                                   ', * will be replaced by the names specified in --chromosomes.'),
                    ('-s/--sample', 'Sample alignments. Use wildcard * to specify multiple files'),
                    ('-c/--control', 'Optional. Use wildcard * to specify multiple files.'
                                     ' Use only this option if you have separate control alignments'),
                    ('-f/--fragment_length', 'Optional. Specify if you know the exact fragment length. '
                                             'Will be estimated if not specified.'),
                    ('-r/--read_length', 'Optional. '),
                    ('-p/--stop_after_p_values', 'Optional. Default False. Set to True in order to'
                                                 'stop the whole run after p-values has been computed.'
                                                 ' Useful if wanting to run multiple chromosomes in parallell, and'
                                                 'waiting before continuing from p-values.'),
                    ('-n/--out_name', 'Optional. Out base name. Prepended to output files.'),
                    ('-u/--unique_reads', 'Optional. Number of unique reads. '
                                          'Found by calling count_unique_reads'),
                    ('-G/--genome_size', 'Optional. Number of base pairs covered by '
                                         'graphs in total (on a linear genome). '
                                         'If not set, will be estimated if running on single graph. '
                                         'Must be set if running on multiple graphs.'),
                    ('-D/--keep_duplicates', 'Optional. Set to True in order to keep '
                                             'duplicate input alignments.'),
                    ('-m/--min_fold_enrichment', 'Optional. Minimum fold enrichment required for '
                                               'candidate peaks when estimating fragment length. Default 5.'),
                    ('-M /--max_fold_enrichment', 'Optional. Maximum fold enrichment required for '
                                               'candidate peaks when estimating fragment length. Default 50.'),

                ],
                'method': run_callpeaks2,
        },

    'callpeaks_whole_genome_from_p_values':
        {
            'help': 'Callpeaks on whole genome from pvalues.',
            'arguments':
                [
                    ('chromosome', 'Specific chromosome to find peaks for.'),
                    ('-d/--data_dir', 'Directory containing graphs.'),
                    ('-n/--out_name', 'Optional. eg experiment1_'),
                    ('-f/--fragment_length', ''),
                    ('-r/--read_length', ''),
                    ('-q/--q_threshold', 'Optional. q-value threshold. Default is 0.05.'),
                    ('-m/--variant_maps_path', 'Optional. Path where variant maps are stored. '
                                               'If set, variant maps will be used to try to improve max paths '
                                               'through subgraphs (better handling of insertions and deletions)')
                ],
            'method': run_callpeaks_whole_genome_from_p_values
        },
    'peaks_to_fasta':
        {
            'help': 'Get sequences for intervals in interval file.',
            'arguments':
                [
                    ('sequence_graph', 'Sequence Graph, e.g. graph.nobg.sequences '
                                       '(created by calling create_ob_graph'),
                    ('intervals_file_name', ''),
                    ('out_file_name', '')
                ],
            'method': peaks_to_fasta
        },
    'linear_peaks_to_fasta':
        {
            'help': 'Converts bed file of peaks to fasta',
            'arguments':
                [
                    ('linear_reads_file_name', 'E.g a macs file, peaks.narrowPeak'),
                    ('fasta_file', 'Reference genome fasta file. Will be used to fetch sequences.'),
                    ('out_file_name', '')
                ],
            'method': linear_peaks_to_fasta
        },
    'linear_peaks_to_fasta_summits':
        {
            'help': 'Converts bed file of peaks to fasta (but keep only summits)',
            'arguments':
                [
                    ('linear_reads_file_name', 'E.g a macs file, peaks.narrowPeak'),
                    ('fasta_file', 'Reference genome fasta file. Will be used to fetch sequences.'),
                    ('out_file_name', ''),
                    ('window', 'Number of bps around summits to keep')
                ],
            'method': linear_peaks_to_fasta
        },
    'create_ob_graph':
        {
            'help': 'Create OffsetBased Graph from vg.',
            'arguments':
                [
                    ('vg_json_file_name', 'Vg json file name '
                                          '(created by running vg view -Vj graph.vg > graph.json'),
                    ('-o/--out_file_name', 'Optional. Will use input file base name if unset.')
                ],
            'method': create_ob_graph
        },
    'create_linear_map':
        {
            'help': 'Create linear map from vg snarls.',
            'requires_graph': True,
            'arguments':
                [
                    ('-o/--out_file_base_name', 'Optional out file name'),
                ],
            'method': create_linear_map_interface
        },
    'split_vg_json_reads_into_chromosomes':
        {
            'help': "Split vg json reads by chromosome.",
            'arguments':
                [
                    ('chromosomes', 'Comma-separated list of chromosomes to split reads into'),
                    ('vg_json_reads_file_name', ''),
                    ('range_files_base_name', 'Base name, e.g. dir, to range files')
                ],
            'method': split_vg_json_reads_into_chromosomes
        },
    'concatenate_sequence_files':
        {
            'help': "Merge multiple *_sequence.fasta files.",
            'arguments':
                [
                    ('chromosomes', 'comma delimted, e.g 1,2,3, used to fetch files of type chr1_sequences.fasta, ...'),
                    ('out_file_name', ''),
                    ('-s/--is_summits', 'Optional. Set to True if input files are *_sequences_summits.fasta.'
                                        'If False or not set, will search for *_sequences.fasta'),
                    ('-f/--use_input_file_pattern', 'Optional. If set, will look for input files'
                                                    'following the pattern. Use [chrom] where chromosome should'
                                                    'be replaced. E.g. peaks_[chrom].fasta')
                ],
            'method': concatenate_sequence_files
        },
    'plot_motif_enrichment':
        {
            'help': "Plots motif enrichments using fimo.",
            'arguments':
                [
                    ('fasta1', ''),
                    ('fasta2', ''),
                    ('meme_motif_file', 'something.meme'),
                    ('out_figure_file_name', ''),
                    ('run_fimo', 'Set to True if fimo has not already been run.'),
                    ('plot_title', 'Title above plot'),
                    ('background_model_file', 'Optional. Background model file for fimo. If not set, fimo will use uniform background.')
                ],
            'method': plot_motif_enrichment
        },
    'analyse_peaks':
        {
            'help': 'Analyse linear peaks and graph peaks.',
            'requires_graph': True,
            'arguments':
                [
                    ('linear_peaks_fasta_file_name', ''),
                    ('graph_peaks_fasta_file_name', ''),
                    ('linear_peaks_fimo_results_file', ''),
                    ('graph_peaks_fimo_results_file', ''),
                    ('linear_path_name', 'Ref of chromosome name. Used to find path in vg json file'),
                    ('graph_chromosome', 'None if not relevant'),
                    ('graph_start', 'Start pos in chromosome of graph.'),
                    ('graph_end', 'End pos in chromosome of graph. 0 if covering whole chromosome')
                ],
            'method': analyse_peaks
        },
    'analyse_peaks_whole_genome':
        {
            'help': "Analyse peak results from whole genome run",
            'arguments':
                [
                    ('chromosomes', 'Comma seaparated list of chromosomes to use'),
                    ('results_dir', 'Directory of result files (should contain fasta files from both linear and graph peak calling'),
                    ('graphs_dir', 'Dir containing obg graphs on form 1.nobg, 2.nobg, ... and vg graphs 1.json, 2.json, ...'),
                    ('out_file', 'Out file base name (file endings for different formats will be appended)'),
                    ('-f/--use_graph_fasta', 'Optional. Specify fasta file to use for graph, format file_[chrom].fasta, where [chrom] will be replaced')               
                ],
            'method': analyse_peaks_whole_genome
        },
    'count_unique_reads':
        {
            'help': 'Count unique reads in vg json alignments.',
            'arguments':
                [
                    ('chromosomes', 'Comma-separated list of chromosomes to use, e.g. 1,2,X,8,Y'),
                    ('graphs_location', 'Will use the graphs *_[chromosome].nobg'),
                    ('reads_base_name', 'Will use files *_[chromosome].json')
                ],
            'method': count_unique_reads_interface
        },
    'find_linear_path':
        {
            'help': 'Find linear path through graph.',
            'requires_graph': True,
            'arguments':
                [
                    ('vg_json_graph_file_name', ''),
                    ('linear_path_name', 'Name of path in the vg graph (typically ref or chromosome name'),
                    ('out_file_name', ''),
                ],
            'method': find_linear_path
        },
    'move_linear_reads_to_graph':
        {
            'help': 'Convert SAM to Intervals on graph.',
            'arguments':
                [
                    ('bed_file_name', ''),
                    ('chromosomes', 'Comma separated list of chromosomes to use'),
                    ('data_dir', 'Directory containing graphs and linear path files'),
                    ('out_files_base_name', '')
                ],
            'method': move_linear_reads_to_graph
        },
    'diffexpr':
        {
            'help': 'Find differentially expressed motif matches',
            'requires_graph': True,
            'arguments':
                [
                    ('test_name', ''),
                    ('fimo_file_name', '')
                ],
            'method': differential_expression
        },
    'check_haplotype':
        {
            'help': 'Check motifs for haplotype',
            'arguments':
                [
                    ('data_folder', ''),
                    ('result_folder', ''),
                    ('chrom', ''),
                    ('interval_name', ''),
                    ('from_peakcollection', 'Optional if given then use intervalcollectionfile')
                ],
            'method': check_haplotype
        },
    'analysis_summary':
        {
            'help': 'Check motifs for haplotype',
            'arguments':
                [
                    ('data_folder', ''),
                    ('result_folder', ''),
                    ('chrom', ''),
                    ('interval_name', ''),
                ],
            'method': get_analysis_summaries
        },

    'overlapping_alignments':
        {
            'help': 'Check motifs for haplotype',
            'arguments':
                [
                    ('data_folder', ''),
                    ('result_folder', ''),
                    ('chromosomes', 'Comma-separated list of chromosomes'),
                    ('interval_name', ''),
                ],
            'method': get_overlapping_alignments
        },
    'haplotype_sequence':
        {
            'help': 'Find haplotype sequence around intervals',
            'arguments':
                [
                    ('data_folder', ''),
                    ('fasta_file', ''),
                    ('result_folder', ''),
                    ('chrom', ''),
                    ('interval_name', ''),
                    ('haplotype', '')
                ],
            'method': get_haplotype_sequence
        },
    'motif_locations':
        {
            'help': 'Check motifs for haplotype',
            'arguments':
                [
                    ('data_folder', ''),
                    ('result_folder', ''),
                    ('interval_name', ''),
                    ('chrom', '')
                ],
            'method': get_motif_locations
        },

    'peaks_to_linear':
        {
            'help': 'Converts graph peaks to linear peaks.',
            'arguments':
                [
                    ('peaks_file_name', "Shuold be a JSON intervalcollection, e.g. one created by callpeaks."),
                    ('linear_path_file_name', "Name of linear path file"),
                    ('chromosome', 'Name of chromosome that will be used when writing the bed file'),
                    ('out_file_name', 'Out file name')
                ],
            'method': peaks_to_linear
        },
    'version':
        {
            'help': 'Prints the current version',
            'arguments': [],
            'method': version
        },
    'estimate_shift':
        {
            'help': 'Estimate shift using one or multiple graphs.',
            'arguments':
                [
                    ('chromosomes', 'Graph base names. Set to empty string if only single '
                                    'graph is being used. If whole-genome, use comma-separated '
                                    'list of chromosomes to use, e.g. 1,2,X,8,Y'),
                    ('ob_graphs_location', 'Location of graph files'),
                    ('sample_reads_base_name', 'Will use files [base_name][chromosome].json'),
                    ('min_fold_enrichment', 'Optional. Minimum fold enrichment when '
                                            'finding candidates. Default is 5.'),
                    ('max_fold_enrichment', 'Optional. Maximum fold enrichment when '
                                            'finding candidates. Default is 50.')
                ],
            'method': shift_estimation
        },
    'get_summits':
        {
            'help': 'Get summit around peaks. Will write to a new file, using input file base name + _summits.fasta',
            'requires_graph': True,
            'arguments':
                [
                    ('peaks_fasta_file', 'Fasta file containing graph peaks.'),
                    ('q_values_base_name', 'Base file name of q values from running '
                                           'the peak caller. Will be peak caller output base name + _qvalues'),
                    ('window_size', 'Optional. Number of basepairs to each side '
                                    'from summit to include. Default is 60.')
                ],
            'method': get_summits
        },
    'get_super_summits':
        {
            'help': 'Get summit around peaks. Will write to a new file, using input file base name + _summits.fasta',
            'requires_graph': True,
            'arguments':
                [
                    ('linear_ref', 'Fasta file containing graph peaks.'),
                    ('peaks_fasta_file', 'Fasta file containing graph peaks.'),
                    ('q_values_base_name', 'Base file name of q values from running '
                                           'the peak caller. Will be peak caller output base name + _qvalues'),
                    ('window_size', 'Optional. Number of basepairs to each side '
                                    'from summit to include. Default is 60.'),
                ],
            'method': get_super_summits
        },
    'project_vg_alignments':
        {
            'help': '',
            'requires_graph': True,
            'arguments':
                [
                    ('alignments_json_file_name', ''),
                    ('linear_path_file_name', ''),
                    ('chromosome', ''),
                    ('out_file_name', 'Writes bed to this file')
                ],
            'method': project_vg_alignments
        },
    'check_pruned_graphs_stats':
        {
            'help': 'Uses vg to check stats of pruned graph vs graphs in graph dir.',
            'arguments':
                [
                    ('graphs_dir', ''),
                    ('chromosomes', 'Comma separated list of chromosomes.'
                                    'Will look for graphs on form graph_dir/[chromosome].pruned.vg')
                ],
            'method': check_pruned_graphs_stats
        },
    'vg_json_alignments_to_intervals':
        {
            'help': 'Reads vg json alignments and converts to interval collection',
            'requires_graph': True,
            'arguments':
                [
                    ('vg_json_file_name', 'Vg json file name'),
                    ('out_file_name', 'Out file name')
                ],
            'method': vg_json_alignments_to_intervals
        },
    'vg_json_alignments_to_fasta':
        {
            'help': 'Reads vg json alignments and converts to interval collection',
            'requires_graph': True,
            'arguments':
                [
                    ('vg_json_file_name', 'Vg json file name'),
                    ('data_dir', ""),
                    ('chrom', ""),
                    ('out_file_name', 'Out file name'),
                ],
            'method': create_alignment_fasta
        },
    'get_intersecting_intervals':
        {
            'help': "Finds intervals in first file that interesects any intervals in second file",
            'requires_graph': True,
            'arguments':
                [
                    ('file1', 'Intervalcollection file 1'),
                    ('file2', 'Intervalcollection file 2'),
                    ('out_file_name', 'File to write intersecting intervals to')
                ],
            'method': get_intersecting_intervals
        },
    'split_peaks_by_chromosome':
        {
            'help': 'Splits a peak intervalcollection file by chromosomes',
            'arguments':
                [
                    ('in_file_name', ''),
                    ('chromosomes', 'Comma separated list of chromosomes to look for'),
                    ('out_file_name_ending', 'Will add *chromosome*_ before this name for each chromosome found.')
                ],
            'method': split_peaks_by_chromosome
        },
    'clean_vcfs':
        {
            'help': "Indexes the vcf files so variants can be looked up by node id",
            'arguments':
                [
                    ('data_folder', 'Folder containing graph files and vcf files'),
                    ('chromosomes', 'Comma separated list of chromosome names corresponding to file names')
                ],
            'method': clean_vcf_wrapper
        }
}


class GraphAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        try:
            new_values = obg.Graph.from_numpy_file(values)
        except FileNotFoundError:
            raise GraphNotFoundException()


        setattr(namespace, self.dest, new_values)
        setattr(namespace, "graph_file_name", values)
        try:
            sequencegraph = obg.SequenceGraph.from_file(values + ".sequences")
            setattr(namespace, "sequence_graph", sequencegraph)
            logging.info("Using sequencegraph %s" % (values + ".sequences"))
        except FileNotFoundError:
            logging.info("No sequencegraph found. Will not use sequencegraph.")
            setattr(namespace, "sequence_graph", None)


def run_argument_parser(args):
    # Create parser
    parser = argparse.ArgumentParser(
        description='Graph peak caller',
        prog='graph_peak_caller',
        formatter_class=lambda prog: argparse.HelpFormatter(prog, max_help_position=50, width=100))

    subparsers = parser.add_subparsers(help='Subcommands')

    for command in interface:

        example = ""
        if "example_run" in interface[command]:
            example = "\nExample: " + interface[command]["example_run"]

        subparser = subparsers.add_parser(
            command,
            help=interface[command]["help"] + example)

        subparser.add_argument('-v', '--verbose',
                               help='Verbosity level. 0 (show only warnings and errors), 1 (show info, '
                                    'warnings and errors), 2 (show everything). Default is 1.', dest='verbose',
                               required=False, metavar='N', const=1, type=int, nargs='?', default=1)

        if 'requires_graph' in interface[command]:
            subparser.add_argument('-g', '--graph', action=GraphAction,
                                   help='Graph file name', dest='graph',
                                   required=True, metavar='file.nobg')

        for argument, help in interface[command]["arguments"]:

            if "/" in argument:
                c = argument.split("/")
                short_command = c[0].strip()
                long_command = c[1].strip()
                assert long_command.startswith("--"), "Long command for %s must start with --" % argument
                assert short_command.startswith("-"), "Short command for %s must start with -" % argument

                required = True
                if "Optional" in help or "optional" in help:
                    required = False

                nargs = None
                if "wildcard" in help:
                    nargs = "+"

                subparser.add_argument(short_command, long_command,
                                       dest=long_command.replace("--", ""), help=help,
                                       required=required, nargs=nargs)
            else:
                subparser.add_argument(argument, help=help)
        subparser.set_defaults(func=interface[command]["method"])


    if len(args) == 0:
        parser.print_help()
        sys.exit(1)
    try:
        args = parser.parse_args(args)
    except GraphNotFoundException:
        logging.error("Specified graph file was not found. Aborting.")
        sys.exit(1)

    set_logging_config(args.verbose)

    if hasattr(args, 'func'):
        try:
            args.func(args)
        except (pyvg.vgobjects.IntervalNotInGraphException, InvalidPileupInterval) as e:
            logging.debug(e)
            logging.error("Found an alignment not compatible with the graph that was used."
                          " Are you sure alignments/intervals are mapped to the same graph that was used?"
                          " Turn on debuging with --verbose 2 to see full log.")
    else:
        parser.help()

if __name__ == "__main__":
    run_argument_parser(sys.argv[1:])



from pyvg.conversion import vg_json_file_to_intervals
import offsetbasedgraph as obg
from graph_peak_caller.haplotyping import HaploTyper
import logging


def graph_to_indexed_interval(graph):
    pass


class LinearFilter:
    def __init__(self, position_tuples, indexed_interval):
        self._nodes = set()
        self._nodes.update(indexed_interval.region_paths)
        self._position_tuples = position_tuples
        self._indexed_interval = indexed_interval

    def find_start_positions(self):
        start_positions = {"+": [],
                           "-": [] }

        logging.info("Mapping graph position to linear positions")
        nodes_in_linear_path = set(self._indexed_interval.region_paths)
        i = 0
        for pos in self._position_tuples:
            direction = "+"
            node = pos.region_path_id
            if node < 0:
                direction = "-"

            if i % 10000 == 0:
                logging.info("%d reads processed" % i)
            i += 1

            if abs(node) not in nodes_in_linear_path:
                continue

            offset = self._indexed_interval.get_offset_at_position(
                pos, direction)
            assert isinstance(offset, int), "Offset is of type %s" % type(offset)
            start_positions[direction].append(offset)

        logging.info("Found in total %d positions" % len(start_positions["+"]))
        return start_positions

    @classmethod
    def from_vg_json_reads_and_graph(cls, json_file_name, graph_file_name):
        logging.info("Reading graph %s" % graph_file_name)
        graph = obg.GraphWithReversals.from_numpy_file(graph_file_name)

        logging.info("Getting indexed interval through graph")
        intervals =  vg_json_file_to_intervals(json_file_name, graph)
        haplotyper = HaploTyper(graph, obg.IntervalCollection(intervals))
        haplotyper.build()
        indexed_interval = haplotyper.get_maximum_interval_through_graph()

        intervals =  vg_json_file_to_intervals(json_file_name, graph)
        positions = (interval.start_position for interval in intervals)

        return cls(positions, indexed_interval)




import numpy as np
import logging
from collections import Counter
from itertools import chain
import offsetbasedgraph as obg
import pyvg
from pyvg.sequences import SequenceRetriever
from pyvg.alignmentcollection import AlignmentCollection
from .peakscomparer import PeaksComparerV2, AnalysisResults
from .manually_classified_peaks import \
    CheckOverlapWithManuallyClassifiedPeaks
from .analyse_peaks import LinearRegion
from .differentialbinding import main, MotifLocation
from .fimowrapper import FimoFile
from ..peakcollection import Peak, PeakCollection
from .nongraphpeaks import NonGraphPeakCollection
from .motifenrichment import plot_true_positives
from ..peakfasta import PeakFasta
from ..mindense import DensePileup
from ..sparsediffs import SparseValues, SparseDiffs
from ..intervals import UniqueIntervals
from .util import create_linear_path
from .genotype_matrix import VariantList
from .haplotype_finder import Main, VariantPrecence
from graph_peak_caller.intervals import UniqueIntervals
from offsetbasedgraph.tracevariants import pipeline_func_for_chromosome,\
    summarize_results, AnalysisResults


def _get_motif_locations(args, chrom):

    graph = obg.Graph.from_file(args.data_folder+chrom+".nobg")
    peaks = PeakCollection.from_fasta_file(
        args.result_folder+chrom+"_sequences_" + args.interval_name + ".fasta", graph)
    peaks = list(peaks)
    fimo = FimoFile.from_file(
        args.result_folder+"fimo_graph_chr"+chrom+"/fimo.txt")
    motif_paths = [MotifLocation.from_fimo_and_peaks(entry, peaks).location
                   for entry in fimo.get_best_entries()]
    print(type(motif_paths[0]))
    PeakCollection(motif_paths).to_file(
        args.result_folder+chrom+"_" + args.interval_name + "_motif_paths.intervalcollection", True)


def get_motif_locations(args):
    print(args.chrom)
    [_get_motif_locations(args, chrom) for chrom in args.chrom.split(",")]


class IntervalDict(obg.IntervalCollection):
    """ Should be moved to obg"""
    def to_file(self, file_name, text_file=False):
        return self.to_text_file(file_name)

    @classmethod
    def from_file(cls, file_name, text_file=False, graph=None):
        return cls.create_generator_from_file(file_name, graph=graph)

    @classmethod
    def create_generator_from_file(cls, file_name, graph=None):
        cur_name = None
        intervals_dict = {}
        f = open(file_name)
        for line in f:
            if line.startswith("#"):
                cur_name = line.strip()[1:]
                intervals_dict[cur_name] = []
                continue
            intervals_dict[cur_name].append(cls.interval_class.from_file_line(
                line, graph=graph))
        f.close()
        return cls(intervals_dict)

    def to_text_file(self, file_name):
        logging.info("Writing to text file %s" % file_name)
        f = open(file_name, "w")

        for name, intervals in self.intervals.items():
            f.write("#" + str(name)+"\n")
            for interval in intervals:
                f.writelines(["%s\n" % interval.to_file_line()])
        f.close()
        return file_name


def summarize_diplotypes(types):
    N = len(types)
    if not N:
        return (0, 0, 0, 0, 0)
    counter = Counter()
    for t in types:
        if len(t):
            if isinstance(t[0], str):
                continue
        counter.update(t)
    if not counter:
        return (0, 0, N, -1, -1)
    most_common = counter.most_common(1)[0]
    remaining_types = [t for t in types if most_common[0] not in t]
    remaining_res = summarize_haplotypes(remaining_types)
    return (most_common[1], remaining_res[0], N, most_common[0], remaining_res[-1])


def summarize_haplotypes(types):
    N = len(types)
    if not N:
        return (0, 0, 0)
    counter = Counter()
    for t in types:
        if len(t):
            if isinstance(t[0], str):
                continue
        counter.update(t)
    if not counter:
        return (0, N, -1)
    most_common = counter.most_common(1)[0]
    return (most_common[1], N, most_common[0])


def get_haplotype_sequence(args):
    name = args.data_folder+args.chrom
    main = Main.from_name(name, args.fasta_file, args.chrom)
    intervals = obg.IntervalCollection.from_file(
        args.result_folder+args.chrom+"_" + args.interval_name + ".intervalcollection", True)
    print(main.get_haplo_sequence_around_intervals(intervals, int(args.haplotype)))


def get_overlapping_alignments(args):
    for chrom in args.chromosomes.split(","):
        logging.info("Running on chromosome %s" % chrom)
        graph = obg.Graph.from_file(args.data_folder+chrom+".nobg")
        base_name = args.result_folder + chrom
        try:
            intervals = PeakCollection.from_fasta_file(base_name + "_sequences_" + args.interval_name+".fasta")
        except:
            intervals = PeakCollection.from_file(base_name + "_" + args.interval_name+".intervalcollection", True)
        alignment_collection = AlignmentCollection.from_file(
            args.result_folder + chrom + "_alignments.pickle", graph)
        peaks_dict = {interval.unique_id:
                      list(UniqueIntervals(
                          alignment_collection.get_alignments_on_interval(interval).values()))
                      for interval in intervals}
        logging.info("Writing interval dict to file")
        IntervalDict(peaks_dict).to_file(
            base_name + "_" + args.interval_name+"_alignments.intevaldict")


def _check_haplotype(args, chrom):
    if args.from_peakcollection == "True":
        filename = args.result_folder + chrom + "_" + args.interval_name + ".intervalcollection"
        interval_dict = {peak.unique_id: [peak] for peak in PeakCollection.from_file(filename, True)}
    else:
        dict_filename = args.result_folder + chrom + "_" + args.interval_name + ".intevaldict"
        interval_dict = IntervalDict.from_file(dict_filename).intervals
    pipeline = obg.tracevariants.pipeline_func_for_chromosome(chrom, args.data_folder)
    results = ((peak_id, pipeline(intervals)) for peak_id, intervals in interval_dict.items())
    with open(args.result_folder + chrom + "_" + args.interval_name + "_diplotypes.tsv", "w") as outfile:
        for result in results:
            outfile.write("%s\t%s\n" % result)


def get_analysis_summaries(args):
    nbins = 100
    def get_motif_match_ids(chrom):
        filename = args.result_folder + "fimo_graph_chr%s/fimo.txt" % chrom
        return [line.split("\t")[2] for line in open(filename) if
                not line.startswith("#")]

    def parse_results(chrom, filterfunc=lambda x, y: True):
        motif_ids = get_motif_match_ids(chrom)
        filename = args.result_folder + chrom + "_" + args.interval_name + "_diplotypes.tsv"
        pairs = ((line.split("\t")[0], eval(line.split("\t")[1])) for line in open(filename)
                 if not line.startswith("#"))
        results = [result for peak_id, result in pairs if filterfunc(peak_id, motif_ids)]
        summary = summarize_results(results, nbins)
        return summary

    def get_summary(filterfunc=lambda x, y: True):
        summaries = [parse_results(chrom, filterfunc) for chrom in args.chrom.split(",")]
        summary = sum(summaries)
        means = summary.copy()
        means[1:] /= means[0]
        return means

    pairs = (("all", lambda x, motif_ids: True),
             ("motif", lambda peak_id, motif_ids: peak_id in motif_ids),
             ("nonmotif", lambda peak_id, motif_ids: peak_id not in motif_ids))
    for name, func in pairs:
        out_file_name = args.result_folder + args.interval_name + "_%s_summary.npz" % name
        means = get_summary(func)
        np.savez(out_file_name, summary=means[:6], haplo_hist=means[6:(nbins+6)], diplo_hist=means[nbins+6:2*nbins+6])

def check_haplotype(args):
    [_check_haplotype(args, chrom) for chrom in args.chrom.split(",")]
    get_analysis_summaries(args)


def check_haplotype_old(args):
    all_reads = args.all_reads == "True" if args.all_reads is not None else True
    strict = args.strict == "True" if args.strict is not None else True
    name = args.data_folder+args.chrom
    VariantPrecence.strict = strict
    # VariantPrecence.accept_ref = all_reads
    main = Main.from_name(name, args.fasta_file, args.chrom)
    motif_paths = obg.IntervalCollection.from_file(
        args.result_folder+args.chrom+"_" + args.interval_name + ".intervalcollection", True)
    graph = obg.Graph.from_file(args.data_folder+args.chrom+".nobg")

    if all_reads:
        alignment_collection = AlignmentCollection.from_file(
            args.result_folder + args.chrom + "_alignments.pickle", graph)
        peaks_dict = {i: list(UniqueIntervals(alignment_collection.get_alignments_on_interval(interval).values()))
                      for i, interval in enumerate(motif_paths)}
        IntervalDict(peaks_dict).to_file(
            args.result_folder + args.chrom + args.interval_name + "_reads.intervaldict")
        # peaks_dict = IntervalDict.from_file(args.result_folder + args.chrom + args.interval_name + "_reads.intervaldict").intervals
        base_name = args.result_folder + args.chrom + "_" + args.interval_name
    else:
        peaks_dict = {i: [interval] for i, interval in enumerate(motif_paths)}
        base_name = args.result_folder + args.chrom + "_" + args.interval_name + "_pure"
    # peaks_dict = IntervalDict.from_file(args.result_folder + args.chrom + "_motif_reads.intervaldict").intervals
    result_dict = main.run_peak_set(peaks_dict)
    if strict:
        base_name += "_strict"
    with open(base_name + ".setsummary", "w") as f:
        summaries = [summarize_haplotypes(v) for v in result_dict.values()]
        successes = sum(s[0] == s[1] for s in summaries)
        f.write("# %s/%s\n" % (successes, len(summaries)))
        for k, v in result_dict.items():
            f.write("%s, %s\n" % (k, summarize_haplotypes(v)))

    with open(base_name + "_dip.setsummary", "w") as f:
        summaries = [summarize_diplotypes(v) for v in result_dict.values()]
        successes = sum(s[0]+s[1] == s[2] for s in summaries)
        f.write("# %s/%s\n" % (successes, len(summaries)))
        for k, v in result_dict.items():
            f.write("%s, %s\n" % (k, summarize_diplotypes(v)))

    lines = ("%s\t%s" % (i, result) for i, results in result_dict.items()
             for result in results)
    # motif_paths = list(sorted(motif_paths, key=lambda x: x.region_paths[0]))
    # haplotypes = main.run_peaks(motif_paths)
    with open(base_name + ".setcoverage", "w") as f:
        for line in lines:
            f.write(line+"\n")
        # for haplotype in haplotypes:
        #     f.write(",".join(str(t) for t in haplotype) + "\n")
        #     if len(haplotype) < 3 and "REF" not in haplotype:
        #         print(haplotype)


def concatenate_sequence_files(args):
    chromosomes = args.chromosomes.split(",")
    out_file_name = args.out_file_name

    out_file_name_json = '.'.join(args.out_file_name.split(".")[:-1]) + ".intervalcollection"

    if args.is_summits == "True":
        file_endings = "_sequences_summits.fasta"
    else:
        file_endings = "_sequences.fasta"

    all_fasta_entries = []
    for chromosome in chromosomes:
        print("Processing chromosome %s" % chromosome)

        if args.use_input_file_pattern is not None:
            assert "[chrom]" in args.use_input_file_pattern, \
                "Use [chrom] to specify where chromosome should be replaced."
            try:
                fasta_file = open(args.use_input_file_pattern.replace("[chrom]", chromosome))
            except FileNotFoundError:
                logging.info("Did not find file matching patterh %s. Does those files exist?" % args.use_input_file_pattern)
                raise
        else:
            logging.info("Guessing file name since use_input_file_pattern is not specified")
            fasta_file = open(chromosome + file_endings)

        for line in fasta_file:
            if line.startswith(">"):
                all_fasta_entries.append([line, None])
            else:
                # This is sequence, add to prev entry
                all_fasta_entries[-1][1] = line
        fasta_file.close()

    peaks = []
    for fasta_entry in all_fasta_entries:
        header = fasta_entry[0].rstrip()
        sequence = fasta_entry[1].rstrip()

        interval_json = "{%s" % header.split(" {")[1]
        interval = Peak.from_file_line(interval_json)
        interval.sequence = sequence
        peaks.append(interval)

    peaks = sorted(peaks, key=lambda s: -s.score)
    out_fasta = open(out_file_name, "w")
    out_intervalcollection = open(out_file_name_json, "w")
    i = 0
    for peak in peaks:
        out_fasta.writelines([">peak%d %s\n" % (i, peak.to_file_line())])
        out_fasta.writelines(["%s\n" % peak.sequence])
        out_intervalcollection.writelines(["%s\n" % peak.to_file_line()])
        i += 1
    out_fasta.close()
    out_intervalcollection.close()

    logging.info("Wrote all peaks in sorted order to %s" % out_file_name)
    logging.info("Wrote all peaks in sorted order to %s" % out_file_name_json)


def split_peaks_by_chromosome(args):
    chromosomes = args.chromosomes.split(",")
    peaks = PeakCollection.from_file(args.in_file_name, text_file=True)

    out_peaks = {chrom: [] for chrom in chromosomes}
    #out_files = {chrom: open(chrom + "_" + args.out_file_name_ending) for chrom in chromosomes}

    for peak in peaks:
        if peak.chromosome in chromosomes:
            out_peaks[peak.chromosome].append(peak)
        else:
            logging.warning("Found peak %s that has a chromosome which is not in list of chromosomes %s" % (peak, chromosomes))

    for chrom in chromosomes:
        PeakCollection(out_peaks[chrom]).to_file(chrom + "_" + args.out_file_name_ending, text_file=True)


def find_linear_path(args):
    graph = args.graph
    vg_graph = pyvg.Graph.from_file(args.vg_json_graph_file_name)
    linear_path = create_linear_path(graph, vg_graph,
                                     path_name=args.linear_path_name,
                                     write_to_file=None)
    linear_path.to_file(args.out_file_name)
    logging.info("Wrote to file %s" % args.out_file_name)


def move_linear_reads_to_graph(args):
    chromosomes = args.chromosomes.split(",")
    chrom_lookup = set(chromosomes)
    graphs = {}
    out_files = {}
    linear_paths = {}
    for chrom in chromosomes:
        linear_paths[chrom] = obg.NumpyIndexedInterval.from_file(
            args.data_dir + "/" + chrom + "_linear_pathv2.interval"
        )
        graphs[chrom] = obg.Graph.from_numpy_file(args.data_dir + "/" + chrom + ".nobg")
        out_files[chrom] = open(args.out_files_base_name + "_" + chrom + ".intervalcollection", "w")

    bedfile = open(args.bed_file_name, "r")
    i = 0
    for line in bedfile:
        if i % 100000 == 0:
            logging.info("%d reads processed" % i)
        i += 1
        line = line.split()
        is_reverse = line[5] == "-"
        chrom = line[0].replace("chr", "")
        if chrom not in chrom_lookup:
            continue
        start = int(line[1])
        end = int(line[2])
        graph_interval = linear_paths[chrom].get_exact_subinterval(start, end)
        graph_interval.graph = graphs[chrom]

        if is_reverse:
            graph_interval = graph_interval.get_reverse()
            assert graph_interval.region_paths[0] < 0
        out_files[chrom].writelines(["%s\n" % graph_interval.to_file_line()])

    for chrom in chromosomes:
        out_files[chrom].close()


def peaks_to_linear(args):
    # Get approximate linear position of peaks using a linear path
    linear_path = obg.NumpyIndexedInterval.from_file(args.linear_path_file_name)
    peaks = PeakCollection.from_file(args.peaks_file_name, text_file=True)
    linear_peaks = peaks.to_approx_linear_peaks(linear_path, args.chromosome)
    linear_peaks.to_bed_file(args.out_file_name)


def get_summits(args):
    graph = args.graph
    qvalues = SparseDiffs.from_sparse_files(args.q_values_base_name)
    #qvalues = SparseValues.from_sparse_files(args.q_values_base_name)
    dense_qvalues = qvalues.to_dense_pileup(size=graph.number_of_basepairs())
    qvalues = DensePileup(graph, dense_qvalues)
    logging.info("Q values fetched")
    peaks = PeakCollection.from_fasta_file(args.peaks_fasta_file, graph)

    if args.window_size is not None:
        window = int(args.window_size)
    else:
        window = 60
        logging.warning("Using default window size %d when cutting peaks around summits." % window)

    peaks.cut_around_summit(qvalues, n_base_pairs_around=window)
    out_file_name = args.peaks_fasta_file.split(".")[0] + "_summits.fasta"
    peaks.to_fasta_file(out_file_name, args.sequence_graph)
    logging.info("Wrote summits to " + out_file_name)

def get_super_summits(args):
    graph = args.graph
    qvalues = SparseDiffs.from_sparse_files(args.q_values_base_name)
    #qvalues = SparseValues.from_sparse_files(args.q_values_base_name)
    dense_qvalues = qvalues.to_dense_pileup(size=graph.number_of_basepairs())
    qvalues = DensePileup(graph, dense_qvalues)
    logging.info("Q values fetched")
    peaks = PeakCollection.from_fasta_file(args.peaks_fasta_file, graph)

    if args.window_size is not None:
        window = int(args.window_size)
    else:
        window = 60
        logging.warning("Using default window size %d when cutting peaks around summits." % window)

    from offsetbasedgraph import NumpyIndexedInterval
    linear_ref = NumpyIndexedInterval.from_file(args.linear_ref)
    peaks.cut_around_summit_super(qvalues, linear_ref, n_base_pairs_around=window)
    out_file_name = args.peaks_fasta_file.split(".")[0] + "_summits.fasta"
    peaks.to_fasta_file(out_file_name, args.sequence_graph)
    logging.info("Wrote summits to " + out_file_name)


def analyse_peaks_whole_genome(args):
    chromosomes = args.chromosomes.split(",")
    from graph_peak_caller.analysis.peakscomparer import AnalysisResults
    results = AnalysisResults()
    for chrom in chromosomes:
        graph = obg.GraphWithReversals.from_numpy_file(
            args.graphs_dir + chrom + ".nobg")
        logging.info("Reading linear path")
        linear_path = obg.NumpyIndexedInterval.from_file(
            args.graphs_dir + chrom + "_linear_pathv2.interval")

        #   try:
        alignments_file_name = args.results_dir + "/" + chrom + "_alignments.pickle"
        logging.info("Looking for alignment collection at %s" % alignments_file_name)
        #alignments = pyvg.alignmentcollection.AlignmentCollection.from_file(
        #    alignments_file_name, graph)
        alignments = None
        #except FileNotFoundError:
        #    logging.warning("Did not find alignment collection. Will analyse without")
        #    alignments = None
        
        if args.use_graph_fasta is not None:
            graph_fasta = args.use_graph_fasta.replace("[chrom]", chrom)
        else:
            graph_fasta = "%s_sequences_summits.fasta" % chrom
        
        analyser = PeaksComparerV2(
            graph,
            args.results_dir + "macs_sequences_chr%s_summits.fasta" % chrom,
            args.results_dir + graph_fasta,
            args.results_dir + "/fimo_macs_chr%s/fimo.txt" % chrom,
            args.results_dir + "/fimo_graph_chr%s/fimo.txt" % chrom,
            linear_path,
            alignments=alignments,
            chromosome=chrom
        )
        results = results + analyser.results

    print(" === Final results for all chromosomes ===")
    print(results)

    results.to_file(args.out_file + ".pickle")
    results.to_csv(args.out_file + ".csv")
    with open(args.out_file + ".txt", "w") as f:
        f.write(str(results))
    logging.info("Wrote results as pickle to %s and as text to %s"
                 % (args.out_file + ".pickle", args.out_file + ".txt"))


def analyse_manually_classified_peaks(args):
    for chromosome in args.chromosomes.split():
        graph_file_name = args.graphs_location + "/" + chromosome + ".nobg"
        graph = obg.GraphWithReversals.from_numpy_file(graph_file_name)
        CheckOverlapWithManuallyClassifiedPeaks.from_graph_peaks_in_fasta(
                graph,
                args.graphs_location + "/" + chromosome + ".json",
                chromosome,
                args.reads_base_name + chromosome + "_sequences.fasta",
                args.regions_file,
                args.manually_classified_peaks_file)


def analyse_peaks(args):
    graph = args.graph

    end = int(args.graph_end)
    if end == 0:
        region = None
    else:
        region = LinearRegion(args.graph_chromosome,
                              int(args.graph_start), end)
    linear_path = obg.NumpyIndexedInterval.from_file(
        args.linear_path_name)
    PeaksComparerV2(
        graph,
        args.linear_peaks_fasta_file_name,
        args.graph_peaks_fasta_file_name,
        args.linear_peaks_fimo_results_file,
        args.graph_peaks_fimo_results_file,
        linear_path,
        region=region)


def differential_expression(args):
    logging.info("Running differential expression.")
    test_name = args.test_name
    fimo_file_name = args.fimo_file_name
    peaks_file_name = "%s_max_paths.intervalcollection" % test_name
    subgraphs_file_name = "%s_sub_graphs.graphs.npz" % test_name
    node_ids_file_name = "%s_sub_graphs.nodeids.npz" % test_name
    graph = args.graph

    subgraphs = np.load(subgraphs_file_name)
    logging.info("Found %d subgraphs" % len(subgraphs.keys()))
    node_ids = np.load(node_ids_file_name)

    res = main(
        FimoFile.from_file(fimo_file_name),
        PeakCollection.from_file(peaks_file_name, True),
        subgraphs,
        node_ids,
        graph)
    retriever = obg.SequenceGraph.from_file(
        args.graph_file_name + ".sequences")
    out_file_name = test_name + "_diffexpr.fasta"
    out_f = open(out_file_name, "w")
    n = 0
    for expr_diff in res:
        n += 1
        main_seq = retriever.get_interval_sequence(expr_diff.main_path)
        out_f.write("> %s %s\n" % (expr_diff.peak_id, expr_diff.main_count))
        out_f.write(main_seq + "\n")
        var_seq = retriever.get_interval_sequence(expr_diff.var_path)
        out_f.write("> %sAlt %s\n" % (expr_diff.peak_id, expr_diff.var_count))
        out_f.write(var_seq + "\n")
    logging.info("Wrote %d lines to %s" % (n, out_file_name))
    out_f.close()


def plot_motif_enrichment(args):
    fasta1 = args.fasta1
    fasta2 = args.fasta2
    meme = args.meme_motif_file

    logging.info("Got argument background model file: %s " % args.background_model_file)

    plot_true_positives(
        [
            ("Graph Peak Caller", fasta1),
            ("MACS2", fasta2)
        ],
        meme,
        plot_title=args.plot_title.replace("ARABIDOPSIS_", ""),
        save_to_file=args.out_figure_file_name,
        run_fimo=args.run_fimo == "True",
        background_model_file=args.background_model_file
    )


def peaks_to_fasta(args):
    logging.info("Getting sequence retriever")
    retriever = obg.SequenceGraph.from_file(args.sequence_graph)
    logging.info("Getting intervals")
    intervals = PeakCollection.create_generator_from_file(
        args.intervals_file_name)
    logging.info("Writing to fasta")
    PeakFasta(retriever).save_intervals(args.out_file_name,
                                        intervals)


def linear_peaks_to_fasta(args):
    collection = NonGraphPeakCollection.from_bed_file(
        args.linear_reads_file_name)
    collection.set_peak_sequences_using_fasta(
        fasta_file_location=args.fasta_file)
    collection.save_to_sorted_fasta(args.out_file_name)
    logging.info("Saved sequences to %s" % args.out_file_name)

    window = None
    if hasattr(args, "window"):
        if args.window is not None:
            window = int(args.window)
            logging.info("Using window size of %d" % window)
        else:
            logging.warning("Not cutting around summit when writing peaks to fasta" % window)

    summits = NonGraphPeakCollection.from_bed_file(
        args.linear_reads_file_name, cut_around_summit=window)
    summits.set_peak_sequences_using_fasta(
        fasta_file_location=args.fasta_file)
    out_name = args.out_file_name.split(".")[0] + "_summits.fasta"
    summits.save_to_sorted_fasta(out_name)
    logging.info("Saved summits to %s" % out_name)



from graph_peak_caller.peakcollection import PeakCollection
import pyvg
import logging
from .util import create_linear_path
import offsetbasedgraph as obg


class CheckOverlapWithManuallyClassifiedPeaks:

    def __init_(self, chromosome, reads, true_peaks_file):
        # Filter out reads that are not in the regions in region_bed_file_name
        self.chromosome = chromosome
        self.reads = reads

    @classmethod
    def from_graph_peaks_in_fasta(cls, graph, vg_graph_json_file_name, chromosome, fasta_file_name, regions_bed_file, true_peaks_file):
        reads = PeakCollection.from_fasta_file(fasta_file_name, graph=graph)
        vg_graph = pyvg.vg.Graph.create_from_file(vg_graph_json_file_name, limit_to_chromosomes=chromosome)
        logging.info("Finding linear path")

        linear_path_file = "linear_path_%s.intervalcollection" % chromosome
        try:
            linear_path = obg.IntervalCollection.from_file(linear_path_file, text_file=True).intervals[0]
            linear_path = linear_path.to_indexed_interval()
        except FileNotFoundError:
            linear_path = create_linear_path(graph, vg_graph, path_name=chromosome, write_to_file=linear_path_file)

        linear_path.graph = graph

        filtered_reads = []

        # Convert regions to intervals in graph
        logging.info("Converting regions to regions in graph")
        graph_regions = []
        bed_file = open(regions_bed_file)
        for line in bed_file:
            print(line)
            line = line.split()
            chr = line[0]
            start = int(line[1])
            end = int(line[2])

            if chr != "chr%s" % chromosome:
                logging.info("Skipping %s, %d, %d" % (chr, start, end))
                continue

            graph_interval = linear_path.get_subinterval(start, end)
            graph_regions.append(graph_interval)

        assert len(graph_regions) > 0, " Found not graph regions for chr %d" % chromosome
        graph_regions = PeakCollection(graph_regions)

        # Filter out reads not overlapping with linear regions
        for read in reads:
            n_overlapping = graph_regions.get_overlapping_intervals(read, minimum_overlap=1)

            if n_overlapping:
                filtered_reads.append(read)

        logging.info("Found %d reads in graph regions" % len(filtered_reads))

        return cls(chromosome, reads, true_peaks_file)



import numpy as np
import matplotlib.pyplot as plt
from numpy.random import normal, multivariate_normal
import subprocess


def generate_tracks(N):
    tracks = np.zeros((2, N*4))
    tracks[:, :N] = multivariate_normal([1, 1], [[1, 1], [1, 1]], N).transpose()
    tracks[0:, N:2*N] = normal(1, 1, (1, N))
    tracks[1:, 2*N:3*N] = normal(1, 1, (1, N))
    return [tracks[0, :], tracks[1, :]]


def create_replicates(track, n_replicates=2, sigma=0.5):
    replicate_tracks = []
    for i in range(0, n_replicates):
        replicate_tracks.append(track + normal(0, sigma, len(track)))

    return replicate_tracks


def write_track_to_bed(track, file_name):
    lines = []

    for i in range(0, len(track)):
        start_position = 20 * i
        end_position = start_position + 10
        score = track[i]
        lines.append("chr1\t%s\t%s\t.\t0\t.\t%.5f\t-1\t%.5f\t5\n" %
                        (start_position, end_position, score, score))

    file = open(file_name, "w")
    file.writelines(lines)
    file.close()


def run_idr(peak_file_name1, peak_file_name2, out_file_name):
    commmand = ["idr", "--quiet", "--samples", peak_file_name1,
                peak_file_name2, "-o", out_file_name]
    ps = subprocess.Popen(commmand, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = ps.communicate()


def forbes(track1, track2):
    observed = sum(track1*track2)
    expected = sum(track1)*sum(track2)/float(len(track1))
    return float(observed)/expected


def idr_bed_to_track_array(bed_file_name, track_length, idr_treshold=540):
    track_array = np.zeros(track_length)
    for line in open(bed_file_name):
        l = line.split()
        start = int(l[1])
        idr_score = int(l[4])
        if idr_score >= idr_treshold:
            #print(idr_score)
            track_position = int(start / 20)
            assert track_position < len(track_array) , \
                "Track position %d from start %d is larger than track length %d" % (track_position, start, len(track_array))
            track_array[track_position] = 1

    return track_array


def noisy_idr(experiments, sigma):
    replicates = [create_replicates(track, sigma=sigma) for track in experiments]

    for name, experiment in zip(["A", "B"], replicates):
        for i, replicate in enumerate(experiment):
            write_track_to_bed(np.exp(replicate), name+str(i) + ".bed")

    for name in ["A", "B"]:
        run_idr(name+"0.bed", name+"1.bed", name + "out.bed")
    final_tracks = {}
    for name in ["A", "B"]:
        final_tracks[name] = idr_bed_to_track_array(name+"out.bed", N*4)

    return forbes(final_tracks["A"], final_tracks["B"]), sum(final_tracks["A"]), sum(final_tracks["B"])


def pipeline(N, sigma):
    experiments = generate_tracks(N)
    return noisy_idr(experiments, sigma)


def test_with_noise(sigma, repeats=100, size=100):
    print("Noise test %.4f" % sigma)
    experiments = generate_tracks(size)
    forbes_values = np.zeros(repeats)
    lengths_A = np.zeros(repeats)
    lengths_B = np.zeros(repeats)
    for i in range(repeats):
        print("repeat %d/%d" % (i, repeats))
        values = noisy_idr(experiments, sigma)
        forbes_values[i] = values[0]
        lengths_A = values[1]
        lengths_B = values[2]

    return np.array([np.mean(forbes_values), np.mean(lengths_A), np.mean(lengths_B)])


N = 20
track_length = N * 4
xs = np.linspace(0.01, 0.4, num=5)
experiment_results = np.array([test_with_noise(x, 400, N) for x in xs])
for i in range(0, len(experiment_results[0])):
    ys = experiment_results[:,i]
    print(ys)
    plt.plot(xs, ys)
    plt.show()



import numpy as np
from .indel_scores import get_end_values_func, get_start_values_func


def max_path_func(pileup, graph, variant_maps):
    not_variant = (variant_maps.snps == 0) & (variant_maps.insertions == 0)
    get_end_value = get_end_values_func(pileup, graph)
    get_start_value = get_start_values_func(pileup, graph)

    def classify_node(prev_node, node):
        if variant_maps.deletions.from_ids[prev_node] & variant_maps.deletions.to_ids[node]:
            return "DEL"
        if variant_maps.insertions[node-graph.min_node]:
            return "INS"
        if variant_maps.snps[node-graph.min_node]:
            return "SNP"
        return "REF"

    def get_max_path(node_ids):
        node_set = set(node_ids)
        node_ids = np.asanyarray(node_ids)
        start_values = dict(zip(node_ids, get_start_value(node_ids-graph.min_node)))
        end_values = dict(zip(node_ids, get_end_value(node_ids-graph.min_node)))

        node_ids = node_ids-graph.min_node
        non_variants = node_ids[not_variant[node_ids]]
        start = np.min(non_variants) if len(non_variants) else np.min(node_ids)
        cur_node = start+graph.min_node
        path = []
        while True:
            path.append(cur_node)
            next_nodes = [node for node in graph.adj_list[int(cur_node)]
                          if node in node_set]
            if not next_nodes:
                break
            var_dict = {"REF": [], "INS": [], "DEL": [], "SNP": []}
            for node in next_nodes:
                var_dict[classify_node(cur_node, node)].append(node)
            best_vars = {key: max((start_values[node], node) for node in nodes) if len(nodes) else (-1, -1)
                         for key, nodes in var_dict.items()}
            current_best = best_vars["REF"]
            if best_vars["SNP"][0] > best_vars["REF"][0]:
                current_best = best_vars["SNP"]
            # current_best = max(best_vars["REF"], best_vars["SNP"])
            if best_vars["INS"][0] >= current_best[0]/2:
                current_best = best_vars["INS"]
            cur_value = end_values[cur_node]
            if current_best[0] <= cur_value/2:
                if current_best[0] <= best_vars["DEL"][0]/2:
                    current_best = best_vars["DEL"]
            cur_node = current_best[1]
        return path
    return get_max_path



from graph_peak_caller.peakcollection import NonGraphPeakCollection, PeakCollection
from offsetbasedgraph import NumpyIndexedInterval, Graph

def macs_to_graph_peaks(folder):
    for chrom in ["1", "2", "3", "4", "5"]:
        path = NumpyIndexedInterval.from_file("/data/bioinf/tair2/" + chrom  + "_linear_pathv2.interval")
        graph = Graph.from_file("/data/bioinf/tair2/" + chrom + ".nobg")
        macs_peaks = PeakCollection.from_fasta_file(folder + "/macs_sequences_chr%s_summits_unique.fasta" % chrom, graph)
        macs_peaks.to_file(folder + "/%s_macs_unique_graph_summits.intervalcollection" % chrom, True)

if __name__ == "__main__":
    import sys
    folder = sys.argv[1]
    macs_to_graph_peaks(folder)



import numpy as np
import matplotlib.pyplot as plt


def plot(base_name):
    def get_hist(s):
        return s["summary"][0]*s["diplo_hist"]
    motif = np.load(base_name + "/limited_summits_alignments_motif_summary.npz")
    nonmotif = np.load(base_name + "/limited_summits_alignments_nonmotif_summary.npz")
    motif_hist = get_hist(motif)
    nonmotif_hist = get_hist(nonmotif)
    cum_motif = np.cumsum(motif_hist)
    cum_nonmotif = np.cumsum(nonmotif_hist)
    cum_total = cum_motif + cum_nonmotif
    ratio = np.where(cum_total == 0, 0, cum_motif/cum_total)
    plt.plot(np.linspace(0, 1, 100), ratio, label=base_name)

if __name__ == "__main__":
    import sys
    paths = sys.argv[1].split(",")
    for path in paths:
        plot(path)
    plt.xlabel("Ratio of reads covered by diplotypes threshold")
    plt.ylabel("Motif match percentage")
    plt.legend()

    plt.show()



import numpy as np
from itertools import chain


class SegmentAnalyzer:
    def __init__(self, segments, node_ids, node_indexes):
        self._segments = segments
        self._node_ids = node_ids
        self.node_indexes = node_indexes
        self.fulls = []
        self.starts = []
        self.ends = []

    def _get_spanning_nodes(self, start_ids, end_ids):
        spanning = end_ids > start_ids+1
        return list(chain.from_iterable(
            range(start+1, end) for start, end in
            zip(start_ids[spanning], end_ids[spanning])))

    def _is_starts(self, offsets, node_ids):
        return offsets == self.node_indexes[node_ids-1]

    def _is_ends(self, offsets, node_ids):
        return offsets == self.node_indexes[node_ids]

    def get_starts(self):
        a = np.transpose(np.array(self.starts))
        if not a.size:
            return np.empty((2, 0), dtype="int")
        a[1] = a[1]-self.node_indexes[a[0]-1]
        return a.astype("int")

    def get_ends(self):
        a = np.transpose(np.array(self.ends, dtype="int"))
        if not a.size:
            return np.empty((2, 0), dtype="int")
        a[1] = self.node_indexes[a[0]]-a[1]
        return a.astype("int")

    def get_fulls(self):
        a = np.array(self.fulls, dtype="int")
        if not a.size:
            return np.empty((2, 0))
        sizes = self.node_indexes[a] - self.node_indexes[a-1]
        return np.vstack((a, sizes.astype("int"))).astype("int")

    def add_starts(self, node_ids, offsets):
        self.starts.extend(zip(node_ids, offsets))

    def add_ends(self, node_ids, offsets):

        self.ends.extend(zip(node_ids, offsets))

    def add_fulls(self, node_ids):
        self.fulls.extend(node_ids)

    def handle_internal(self):
        is_starts = self._is_starts(self._internal_segments[:, 0],
                                    self._internal_ids[:, 0])
        is_ends = self._is_ends(self._internal_segments[:, 1],
                                self._internal_ids[:, 1])
        full_mask = is_starts & is_ends
        self.add_fulls(self._internal_ids[:, 0][full_mask])
        start_mask = is_starts & ~full_mask
        self.add_starts(self._internal_ids[:, 1][start_mask],
                        self._internal_segments[:, 1][start_mask])
        end_mask = is_ends & ~full_mask
        self.add_ends(self._internal_ids[:, 0][end_mask],
                      self._internal_segments[:, 0][end_mask])
        self.internals = self._internal_segments[~is_starts & ~is_ends]

    def handle_multinode(self):
        self.add_fulls(self._get_spanning_nodes(self._spanning_ids[:, 0],
                                                self._spanning_ids[:, 1]))
        is_starts = self._is_starts(self._spanning_segments[:, 0],
                                    self._spanning_ids[:, 0])
        is_ends = self._is_ends(self._spanning_segments[:, 1],
                                self._spanning_ids[:, 1])
        self.add_fulls(self._spanning_ids[:, 0][is_starts])
        self.add_fulls(self._spanning_ids[:, 1][is_ends])
        self.add_ends(self._spanning_ids[:, 0][~is_starts],
                      self._spanning_segments[:, 0][~is_starts])
        self.add_starts(self._spanning_ids[:, 1][~is_ends],
                        self._spanning_segments[:, 1][~is_ends])

    def run(self):
        multinodes = self._node_ids[:, 0] != self._node_ids[:, 1]
        self._internal_segments = self._segments[~multinodes]
        self._internal_ids = self._node_ids[~multinodes]
        self._spanning_segments = self._segments[multinodes]
        self._spanning_ids = self._node_ids[multinodes]
        self.multinodes = multinodes
        self.handle_multinode()
        self.handle_internal()


class SegmentSplitter(SegmentAnalyzer):
    def handle_internal(self):
        is_starts = self._is_starts(self._internal_segments[:, 0],
                                    self._internal_ids[:, 0])
        is_ends = self._is_ends(self._internal_segments[:, 1],
                                self._internal_ids[:, 1])
        self.full_mask = is_starts & is_ends
        self.start_mask = is_starts & ~self.full_mask
        self.end_mask = is_ends & ~self.full_mask
        self.internal_mask = ~is_starts & ~is_ends

    def handle_multinode(self):
        spanned = self._get_spanning_nodes(self._spanning_ids[:, 0],
                                           self._spanning_ids[:, 1])
        n_spanned = len(spanned)
        n_old = self._segments.shape[0]
        n_new = self._spanning_segments.shape[0]
        n_total = n_spanned + n_old + n_new
        splitted_segments = np.empty((n_total, 2), dtype="int")
        if n_old:
            splitted_segments[:n_old] = self._segments
            splitted_segments[:n_old, 1][self.multinodes] = self.node_indexes[
                self._spanning_ids[:, 0]]
        splitted_segments[n_old:n_old+n_new, 0] = self.node_indexes[
            self._spanning_ids[:, 1]-1]
        splitted_segments[n_old:n_old+n_new, 1] = self._spanning_segments[:, 1]
        spanned = np.array(spanned, dtype="int")
        if n_spanned:
            splitted_segments[-n_spanned:, 0] = self.node_indexes[spanned-1]
            splitted_segments[-n_spanned:, 1] = self.node_indexes[spanned]
        self.splitted_segments = splitted_segments
        self._internal_segments = splitted_segments
        ids = np.r_[self._node_ids[:, 0],
                    self._node_ids[:, 1][self.multinodes],
                    spanned]
        # assert np.all(ids >= self._graph.blocks.node_id_offset), (self._graph.blocks.node_id_offset,
        #                                                           ids[ids<self._graph.blocks.node_id_offset])
        self._internal_ids = np.hstack((ids[:, None], ids[:, None]))


def test_segments_cleaner():
    offsets = np.array([80, 100,
                        180, 220,
                        240, 250,
                        300, 400,
                        500, 520,
                        610, 810])

    offsets = offsets.reshape((offsets.size//2, 2))
    node_ids = np.empty_like(offsets)
    node_ids[:, 0] = offsets[:, 0] // 100
    node_ids[:, 1] = ((offsets[:, 1]-1) // 100)
    node_ids += 1
    node_indexes = np.arange(0, 1201, 100)
    segments = SegmentsAnalyzer(offsets, node_ids, node_indexes)
    segments = np.unique(np.sort(segments.ravel()))

if __name__ == "__main__":
    test_segments_cleaner()
    



from setuptools import setup

setup(name='graph_peak_caller',
      version='1.2.1',
      description='Graph peak caller',
      url='http://github.com/uio-bmi/graph_peak_caller',
      author='Ivar Grytten and Knut Rand',
      author_email='',
      license='MIT',
      packages=['graph_peak_caller', 'graph_peak_caller.legacy', 'graph_peak_caller.control',
                'graph_peak_caller.sample', 'graph_peak_caller.postprocess',
                'graph_peak_caller.analysis', "graph_peak_caller.shiftestimation"],
      zip_safe=False,
      install_requires=['pymysql', 'numpy', 'filecache', 'scipy',
                        'memory_profiler', 'python-coveralls', 'matplotlib',
                        'biopython', 'pyfaidx', 'pyvg', 'offsetbasedgraph'
                        ],
      classifiers=[
            'Programming Language :: Python :: 3'
      ],
      entry_points = {
        'console_scripts': ['graph_peak_caller=graph_peak_caller.command_line_interface:main'],
      }

      )

"""
To update package:
#Update version number manually in this file

sudo python3 setup.py sdist
sudo python3 setup.py bdist_wheel
twine upload dist/graph_peak_caller-1.1.1.tar.gz
"""



from offsetbasedgraph.graphtraverser import GraphTraverser
from collections import defaultdict

class GraphIndex(object):
    def __init__(self, index):
        self.index = index

    def create_from_graph(self, graph, length=100):
        self._create_index(graph, length)

    def get_node(self, node):
        yield (node, 0)
        for child in self.index[node]:
            yield child

    def _create_index(self, graph, length):
        self.index = defaultdict(list)
        for node in graph:
            self._traverse_from_node(graph, node, length)
            self._traverse_from_node(graph, -node, length)

    def _traverse_from_node(self, graph, node, length):
        if node > 0:
            traverser = GraphTraverser(graph, direction=1)
        else:
            traverser = GraphTraverser(graph, direction=-1)

        visited = defaultdict(int)
        for next_node in traverser.adj_list[node]:
            traverser.extend_from_block(next_node, length, visited)

        for to_node, to_length in visited:
            self.index[node].append((to_node, to_length))

    @classmethod
    def from_file(cls, file_name):
        pass

    def to_file(self, file_name):
        pass


class GraphExtender(object):
    def __init__(self, graph_index):
        self.index = graph_index

    def extend_from_position(self, node, offset, extension_length):
        return ((node, length-offset) for node, length in
                self.index.get_node(node) if length-offset < extension_length)


class DensePileupExtender(object):
    def __init__(self, graph_extender, dense_pileup):
        self.graph_extender = graph_extender
        self.pileup = dense_pileup

    def extend_from_position(self, node, offset, extension_length):
        extensions = self.graph_extender.extend_from_position(node, offset, extension_length)
        for extension in extensions:
            print("Extension: %d,%d" % (extension))
            to_node, length_to_start = extension
            to_node_end = extension_length - length_to_start
            to_node_start = max(0, -length_to_start)
            print("start, end: %d, %d" % (to_node_start, to_node_end))
            yield self.pileup.data.node_range_to_value_indexes(to_node, to_node_start, to_node_end)



"""

TODO:
    - Create graph index from graph
    - Create sample extender

For every node:

    List of nodes with start less than length away from node start
        For every node in this list:
            Contain start and end position in dense pileup values


    Given position in graph, will ask for all possible positions in dense pileup that are n basepairs (extension_size) away
    Position: node, offset
        1) Ask node for list of children
        2) For every node in children
            1) Get start, end densepileup indexes
            2) Adjust start end to:
                n_start = max(end, start + offset)
                n_end = min(n_start + extension_size, end)
"""



